import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio

sys.path.append('C:\\Users\\Public\\repos\\xdart')

from xdart.modules.pySSRL_bServer.bServer.BL_Error import *


class BLCommand():
    """A class to keep track of each beamline interaction: a command and (potentially) response.  
    Keep track of whether there was an error and keep track of the list of responses 

    I plan to keep track of the counterUser variable in here so I can verify 

    """
    def __init__(self, beamline, command, sendterminator=None, readterminator=None, listSeparator=None, needsResponse=False, pauseAfterWrite=None, pauseAfterRead=None):

        self.beamline = beamline
        self.command = copy.copy(command)

        self.needsResponse = needsResponse is True
        self.response = []

        self.readterminator = copy.copy(readterminator)
        self.sendterminator = copy.copy(sendterminator)
        self.listSeparator = copy.copy(listSeparator)

        if pauseAfterWrite is not None:
            self.pauseAfterWrite = copy.copy(pauseAfterWrite)
        else:
            self.pauseAfterWrite = 0
        
        
        if pauseAfterRead is not None:
            self.pauseAfterRead = copy.copy(pauseAfterRead)
        else:
            self.pauseAfterRead = 0
        
        self.writeSuccess = False
        self.readSuccess = False

        self.senderror = False
        self.readerror = False
        self.fatalerror = False

        self.timeCreated = time.time()


        
    def __del__(self):
        pass
    

    async def execute(self):
        """This function should execute a single command and read the value if required
        (basically, go through and execute everything necessary for the beamline interaction)"""

        #print("executing cmd")
        #Keep the commands in the queue
        #self.beamline.completedCommands.append(self)

        try: #Try to execute the command
            debug("execute: '{}'".format(self.command))
            #print('try')
            await self.beamline.writeCommand(self.command, terminator=self.sendterminator)
            #print('sent')
        except: #If we pick up an error, return the error and set the send error flag
            warning("Problem writing command")
            self.senderror = True
            raise
            
        else: #If the send goes well, set the write suceess flag and check to see if we need a response
            self.writeSuccess = True
            self.timeWritten = time.time()
            debug("execute: command written")

            try: #Wait after write if this has been requested
                await asyncio.sleep(self.pauseAfterWrite)
            except:
                debug("execute: error waiting")
                raise

            
            if self.needsResponse is True: #If a response is needed
                try: #Try to read the response
                    debug("execute: Reading response")
                    self.response = await self.beamline.readResponse(terminator = self.readterminator, listSeparator = self.listSeparator)
                except: #If we get an error, we should do something about it. For now lets just pass it
                    debug("")
                    self.readerror = True
                    raise
                else: #If the read goes well, set 
                    debug("execute: response read")
                    self.readSuccess = True
                    self.timeRead = time.time()
                    
                    try: #do a pause after the read if requested
                        await asyncio.sleep(self.pauseAfterRead)
                    except:
                        pass
                
            
        return

    def toSequence(self):
        """Create a command sequence containing only this command. This will be very useful for sequence addition"""
        newSequence = BL_CommandSequence(singleCommand=self)

        return newSequence
    

    def __repr__(self):
        """Representation of the current object"""
        res = []
        
        res.append("Command: {}".format(self.command))
        
        res.append("Response: {} items".format(len(self.response)))
        for m in self.response:
            res.append(" '{}'".format(m))
            
        return("\n".join(res))

    def typecastResponse(self, typecastList):
        """Return the typecast form of the response. This is a little lazy. Try harder..."""
        #print("response: {}".format(self.response))
        if len(typecastList) != len(self.response): #This is a pretty basic error check.... if the beamline doesn't return the right number of parameters, we should throw a beamline error and let it try to recover or quit.
            raise(BLCommandError('typecastResponse', "cmd '{}' response '{}' typecastList '{}' ".format(self.command, self.response, typecastList)))

        converted_response = []
        for ctr in range(len(typecastList)):
            current_response = self.response[ctr]
            current_typecast = typecastList[ctr]

#            print(current_response)
#            print(current_typecast(current_response))

            if current_typecast == bool: #Turns out bool("0") is True... in fact any string returns true. Convert to integer and then to bool
                converted_response.append( bool(int(current_response)))
            else:
                converted_response.append(current_typecast(current_response))


        #print(converted_response)
        return converted_response
        
        
    
    def __add__(self, nextSeq): 
        """Added commands must be converted to sequences. This routine gets called if a command is on the left
        and prevents python from messing up the order of operations."""
        newSequence = BL_CommandSequence(singleCommand=self)
        
        #Once the first command has been converted to a sequence, we can call the normal CommandSequence version of add
        newSequence = newSequence + nextSeq 
        
        return newSequence
                    




class BL_CommandSequence():
    def __init__(self, singleCommand = None):
        
        self.commandList = []
        
        if singleCommand is not None:
            self.commandList.append(singleCommand)
        
        self.current = 0
        
        return
    
    
    def __add__(self, nextSeq): 
        """Add a command sequence to a command sequence or a single command with 'self' preceeding 'nextSeq' in the 
        list. While order of operations matters, we have overloaded the 'add' function of the LMD_command class so 
        we are guaranteed to always have the correct ordering"""
        
        newSequence = BL_CommandSequence()

        
        try: # If a command was passed as a CommandSequence, lets convert it to a sequence with only this command
            nextSeq = nextSeq.toSequence()
        except: #If there is an exception, we just keep moving on
            pass
        
        for cCmd in self.commandList: #Add the 'self' sequence first
            newSequence.commandList.append(cCmd) 
        
        for cCmd in nextSeq.commandList: #Add the 'nextSeq' after this
            newSequence.commandList.append(cCmd)
        
        return newSequence
        

#    def __radd__(self, aCmd): #This is not necessary because we put in an __add__ for the LMD_Command. We should never need to wory about operator presecend

    
    def __iadd__(self, nextSeq):
        """An inplace add will convert a command to a sequence if necessary and then append it to the end of a current object"""
        try:
            nextSeq = nextSeq.toSequence()
        except:
            pass
        
        for cCmd in nextSeq.commandList:
            self.commandList.append(cCmd)
        
        return(self)

    
    
    def __getitem__(self, index):

        try:
            rCmd = self.commandList[index]
        except IndexError:
            raise
        else:
            return rCmd

    def __setitem__(self, index, value):
        try:
            self.commandList[index] = value
        except IndexError:
            raise
    
    def __delitem__(self, index):
        try:
            self.commandList.remove(index)
        except IndexError:
            raise
        
    def __len__(self):
        return( len(self.commandList))
    
    def __repr__(self):
        """Representation of the current object"""
        res = []
        
        for m in self.commandList:
            res.append(m.__repr__())
            
        return("\n".join(res))
    
    async def execute(self):
        """Execute a sequence of commands"""
        for currentCommand in self.commandList:
            try:
                await currentCommand.execute()
            except:
                print("error in command.execute: {}".format(sys.exc_info()[0]))
                raise
            else:
                pass

        return




from pyFAI.detectors import Detector
from pyFAI import detector_factory
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import pygix, pyFAI
import yaml
import numpy as np


class PONI(object):
    """Container for values needed in AzimuthalIntegrator of pyFAI.
    Also provides some utility functions for reading and writing data
    
    attributes:
        detector: pyFAI Detector, detector object
        dist: float, distance to detector, meters
        poni1: float, location of point of nearest incidence 1, meters
        poni2: float, location of point of nearest incidence 2, meters
        rot1: float, angle of rotation 1, radians
        rot2: float, angle of rotation 2, radians
        rot3: float, angle of rotation 3, radians
        wavelength: float, wavelength of energy, used for
            collecting data, meters
    """
    _poni_keys = {
        'Distance': 'dist',
        'Poni1': 'poni1',
        'Poni2': 'poni2',
        'Rot1': 'rot1',
        'Rot2': 'rot2',
        'Rot3': 'rot3',
        'Wavelength': 'wavelength'
    }

    def __init__(self, dist=0, poni1=0, poni2=0, rot1=0, rot2=0, rot3=0, 
                 wavelength=1e-10, detector=Detector(100e-6, 100e-6)):
        """
        dist: float, distance to detector, meters
        poni1: float, location of point of nearest incidence 1, meters
        poni2: float, location of point of nearest incidence 2, meters
        rot1: float, angle of rotation 1, radians
        rot2: float, angle of rotation 2, radians
        rot3: float, angle of rotation 3, radians
        wavelength: float, wavelength of energy, used for
            collecting data, meters
        detector: pyFAI Detector, detector object
        """
        self.dist = dist
        self.poni1 = poni1 
        self.poni2 = poni2 
        self.rot1 = rot1 
        self.rot2 = rot2 
        self.rot3 = rot3 
        self.wavelength = wavelength
        self.detector = detector

    def to_dict(self):
        """Converts all attributes to dictionary with pyFAI names
        """
        return {
            'Distance': self.dist,
            'Poni1': self.poni1,
            'Poni2': self.poni2,
            'Rot1': self.rot1,
            'Rot2': self.rot2,
            'Rot3': self.rot3,
            'Wavelength': self.wavelength,
            'Detector': self.detector.name,
            'Detector_config': {
                'pixel1': self.detector.pixel1,
                'pixel2': self.detector.pixel2,
                'max_shape': self.detector.max_shape
            }
        }

    @classmethod
    def from_dict(cls, input):
        """Creates PONI object from a dictionary.
        """
        out = cls()
        for key in input:
            setattr(out, key, input[key])
        return out
    
    @classmethod
    def from_yamdict(cls, input):
        """Creates a PONI object from the yaml dictionary returned
        from reading a .poni file
        """
        out = cls()
        for key in input:
            if key == 'Detector':
                if 'Detector_config' in input:
                    if 'max_shape' in input['Detector_config']:
                        if not (input['Detector_config']['max_shape'] is None or 
                                type(input['Detector_config']['max_shape']) == str):
                            input['Detector_config']['max_shape'] = list(input['Detector_config']['max_shape'])
                    out.detector = detector_factory(
                        input['Detector'], config=input['Detector_config']
                    )
                elif 'PixelSize1' in input and 'PixelSize2' in input:
                    out.detector = detector_factory(
                        input['Detector'], 
                        config={
                            'pixel1': input['PixelSize1'],
                            'pixel2': input['PixelSize2'],
                        }
                    )
                else:
                    out.detector = detector_factory(input['Detector'])
            elif key == 'Wavelength':
                if type(input[key]) == str:
                    out.wavelength = eval(input[key])
                else:
                    out.wavelength = input[key]
            else:
                try:
                    setattr(out, cls._poni_keys[key], input[key])
                except KeyError:
                    pass
        return out
    
    @classmethod
    def from_yaml(cls, stream):
        """Creates a PONI object from a yaml stream.
        """
        input = yaml.safe_load(stream)
        return cls.from_yamdict(input)
    
    @classmethod
    def from_ponifile(cls, file):
        """Creates a PONI object from a .poni file.
        """
        if type(file) == str:
            with open(file, 'r') as f:
                out = cls.from_yaml(f)
        else:
            out = cls.from_yaml(file)
        return out


def get_poni_dict(poni_file):
    """ Read Poni File and convert to Dictionary"""
    ai = pyFAI.load(poni_file)
    poni_keys = ['_dist', '_rot1', '_rot2', '_rot3', '_poni1', '_poni2', 'detector', '_wavelength']

    try:
        poni_dict = {k: ai.__getattribute__(k) for k in poni_keys}
        return poni_dict
    except KeyError:
        return None


def create_ai_from_dict(poni_dict, gi=False):
    """Create Azimuthal Integrator object from Dictionary"""
    ai = AzimuthalIntegrator()
    for k, v in poni_dict.items():
        ai.__setattr__(k, v)

    if not gi:
        if 'MX225' in ai.detector.name:
            ai._rot3 -= np.deg2rad(90)
    else:
        calib_pars = dict(
            dist=ai._dist, poni1=ai._poni1, poni2=ai._poni2,
            rot1=ai._rot1, rot2=ai._rot2, rot3=ai._rot3,
            wavelength=ai._wavelength, detector=ai.detector)
        ai = pygix.Transform(**calib_pars)
        ai.sample_orientation = 3  # 1 is horizontal, 2 is vertical

    return ai



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'integratorUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1440, 402)
        Form.setMaximumSize(QtCore.QSize(16777215, 430))
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame1D = QtWidgets.QFrame(Form)
        self.frame1D.setMaximumSize(QtCore.QSize(16777215, 190))
        self.frame1D.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame1D.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame1D.setLineWidth(20)
        self.frame1D.setMidLineWidth(5)
        self.frame1D.setObjectName("frame1D")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame1D)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.frame1D_layout = QtWidgets.QVBoxLayout()
        self.frame1D_layout.setSpacing(0)
        self.frame1D_layout.setObjectName("frame1D_layout")
        self.frame1D_header = QtWidgets.QFrame(self.frame1D)
        self.frame1D_header.setMaximumSize(QtCore.QSize(16777215, 50))
        self.frame1D_header.setAutoFillBackground(True)
        self.frame1D_header.setStyleSheet("")
        self.frame1D_header.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame1D_header.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame1D_header.setObjectName("frame1D_header")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.frame1D_header)
        self.horizontalLayout_4.setContentsMargins(15, 3, 5, 3)
        self.horizontalLayout_4.setSpacing(0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label1D = QtWidgets.QLabel(self.frame1D_header)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label1D.sizePolicy().hasHeightForWidth())
        self.label1D.setSizePolicy(sizePolicy)
        self.label1D.setMinimumSize(QtCore.QSize(70, 0))
        self.label1D.setMaximumSize(QtCore.QSize(70, 16777215))
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label1D.setFont(font)
        self.label1D.setAutoFillBackground(False)
        self.label1D.setStyleSheet("background-color: rgba(195, 195, 195, 200);")
        self.label1D.setAlignment(QtCore.Qt.AlignCenter)
        self.label1D.setObjectName("label1D")
        self.horizontalLayout_3.addWidget(self.label1D)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.axis1D = QtWidgets.QComboBox(self.frame1D_header)
        self.axis1D.setObjectName("axis1D")
        self.axis1D.addItem("")
        self.horizontalLayout_3.addWidget(self.axis1D)
        spacerItem1 = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.label_npts_1D = QtWidgets.QLabel(self.frame1D_header)
        self.label_npts_1D.setMaximumSize(QtCore.QSize(40, 16777215))
        self.label_npts_1D.setIndent(-1)
        self.label_npts_1D.setObjectName("label_npts_1D")
        self.horizontalLayout_3.addWidget(self.label_npts_1D)
        self.npts_1D = QtWidgets.QLineEdit(self.frame1D_header)
        self.npts_1D.setMaximumSize(QtCore.QSize(55, 16777215))
        self.npts_1D.setInputMethodHints(QtCore.Qt.ImhDigitsOnly)
        self.npts_1D.setObjectName("npts_1D")
        self.horizontalLayout_3.addWidget(self.npts_1D)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_3)
        self.frame1D_layout.addWidget(self.frame1D_header)
        self.frame1D_range = QtWidgets.QFrame(self.frame1D)
        self.frame1D_range.setMaximumSize(QtCore.QSize(16777215, 71))
        self.frame1D_range.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame1D_range.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame1D_range.setObjectName("frame1D_range")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame1D_range)
        self.horizontalLayout_2.setContentsMargins(5, 5, 5, 5)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.gridLayout_1D = QtWidgets.QGridLayout()
        self.gridLayout_1D.setHorizontalSpacing(6)
        self.gridLayout_1D.setObjectName("gridLayout_1D")
        self.radial_autoRange_1D = QtWidgets.QCheckBox(self.frame1D_range)
        self.radial_autoRange_1D.setChecked(True)
        self.radial_autoRange_1D.setObjectName("radial_autoRange_1D")
        self.gridLayout_1D.addWidget(self.radial_autoRange_1D, 0, 4, 1, 1)
        self.unit_1D = QtWidgets.QComboBox(self.frame1D_range)
        self.unit_1D.setMaximumSize(QtCore.QSize(90, 16777215))
        self.unit_1D.setObjectName("unit_1D")
        self.unit_1D.addItem("")
        self.unit_1D.addItem("")
        self.gridLayout_1D.addWidget(self.unit_1D, 0, 0, 1, 1)
        self.label_to1 = QtWidgets.QLabel(self.frame1D_range)
        self.label_to1.setMaximumSize(QtCore.QSize(100, 30))
        self.label_to1.setObjectName("label_to1")
        self.gridLayout_1D.addWidget(self.label_to1, 1, 2, 1, 1)
        self.azim_low_1D = QtWidgets.QLineEdit(self.frame1D_range)
        self.azim_low_1D.setObjectName("azim_low_1D")
        self.gridLayout_1D.addWidget(self.azim_low_1D, 1, 1, 1, 1)
        self.radial_low_1D = QtWidgets.QLineEdit(self.frame1D_range)
        self.radial_low_1D.setObjectName("radial_low_1D")
        self.gridLayout_1D.addWidget(self.radial_low_1D, 0, 1, 1, 1)
        self.azim_autoRange_1D = QtWidgets.QCheckBox(self.frame1D_range)
        self.azim_autoRange_1D.setChecked(True)
        self.azim_autoRange_1D.setObjectName("azim_autoRange_1D")
        self.gridLayout_1D.addWidget(self.azim_autoRange_1D, 1, 4, 1, 1)
        self.azim_high_1D = QtWidgets.QLineEdit(self.frame1D_range)
        self.azim_high_1D.setObjectName("azim_high_1D")
        self.gridLayout_1D.addWidget(self.azim_high_1D, 1, 3, 1, 1)
        self.label_to2 = QtWidgets.QLabel(self.frame1D_range)
        self.label_to2.setObjectName("label_to2")
        self.gridLayout_1D.addWidget(self.label_to2, 0, 2, 1, 1)
        self.label_azim_1D = QtWidgets.QLabel(self.frame1D_range)
        self.label_azim_1D.setMinimumSize(QtCore.QSize(55, 0))
        self.label_azim_1D.setMaximumSize(QtCore.QSize(100, 50))
        self.label_azim_1D.setAlignment(QtCore.Qt.AlignCenter)
        self.label_azim_1D.setObjectName("label_azim_1D")
        self.gridLayout_1D.addWidget(self.label_azim_1D, 1, 0, 1, 1)
        self.radial_high_1D = QtWidgets.QLineEdit(self.frame1D_range)
        self.radial_high_1D.setObjectName("radial_high_1D")
        self.gridLayout_1D.addWidget(self.radial_high_1D, 0, 3, 1, 1)
        self.horizontalLayout_2.addLayout(self.gridLayout_1D)
        self.frame1D_layout.addWidget(self.frame1D_range)
        self.frame1D_buttons = QtWidgets.QFrame(self.frame1D)
        self.frame1D_buttons.setMaximumSize(QtCore.QSize(16777215, 71))
        self.frame1D_buttons.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame1D_buttons.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame1D_buttons.setObjectName("frame1D_buttons")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.frame1D_buttons)
        self.horizontalLayout_6.setContentsMargins(5, 5, 5, 32)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setSpacing(6)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.integrate1D = QtWidgets.QPushButton(self.frame1D_buttons)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.integrate1D.sizePolicy().hasHeightForWidth())
        self.integrate1D.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        self.integrate1D.setFont(font)
        self.integrate1D.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.integrate1D.setObjectName("integrate1D")
        self.horizontalLayout_5.addWidget(self.integrate1D)
        self.advanced1D = QtWidgets.QPushButton(self.frame1D_buttons)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced1D.sizePolicy().hasHeightForWidth())
        self.advanced1D.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        self.advanced1D.setFont(font)
        self.advanced1D.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.advanced1D.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.advanced1D.setObjectName("advanced1D")
        self.horizontalLayout_5.addWidget(self.advanced1D)
        self.horizontalLayout_6.addLayout(self.horizontalLayout_5)
        self.frame1D_layout.addWidget(self.frame1D_buttons)
        self.horizontalLayout.addLayout(self.frame1D_layout)
        self.verticalLayout.addWidget(self.frame1D)
        self.frame2D = QtWidgets.QFrame(Form)
        self.frame2D.setMaximumSize(QtCore.QSize(16777215, 190))
        self.frame2D.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame2D.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2D.setLineWidth(20)
        self.frame2D.setMidLineWidth(5)
        self.frame2D.setObjectName("frame2D")
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout(self.frame2D)
        self.horizontalLayout_12.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_12.setSpacing(0)
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.frame2D_layout = QtWidgets.QVBoxLayout()
        self.frame2D_layout.setSpacing(0)
        self.frame2D_layout.setObjectName("frame2D_layout")
        self.frame2D_header = QtWidgets.QFrame(self.frame2D)
        self.frame2D_header.setMaximumSize(QtCore.QSize(16777215, 50))
        self.frame2D_header.setAutoFillBackground(True)
        self.frame2D_header.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame2D_header.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2D_header.setObjectName("frame2D_header")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout(self.frame2D_header)
        self.horizontalLayout_7.setContentsMargins(15, 3, 5, 3)
        self.horizontalLayout_7.setSpacing(0)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.label2D = QtWidgets.QLabel(self.frame2D_header)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label2D.sizePolicy().hasHeightForWidth())
        self.label2D.setSizePolicy(sizePolicy)
        self.label2D.setMinimumSize(QtCore.QSize(70, 0))
        self.label2D.setMaximumSize(QtCore.QSize(70, 16777215))
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label2D.setFont(font)
        self.label2D.setAutoFillBackground(False)
        self.label2D.setStyleSheet("background-color: rgba(195, 195, 195, 200);")
        self.label2D.setAlignment(QtCore.Qt.AlignCenter)
        self.label2D.setObjectName("label2D")
        self.horizontalLayout_8.addWidget(self.label2D)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem2)
        self.axis2D = QtWidgets.QComboBox(self.frame2D_header)
        self.axis2D.setObjectName("axis2D")
        self.axis2D.addItem("")
        self.axis2D.addItem("")
        self.horizontalLayout_8.addWidget(self.axis2D)
        spacerItem3 = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem3)
        self.label_npts_2D = QtWidgets.QLabel(self.frame2D_header)
        self.label_npts_2D.setMaximumSize(QtCore.QSize(40, 16777215))
        self.label_npts_2D.setIndent(-1)
        self.label_npts_2D.setObjectName("label_npts_2D")
        self.horizontalLayout_8.addWidget(self.label_npts_2D)
        self.npts_radial_2D = QtWidgets.QLineEdit(self.frame2D_header)
        self.npts_radial_2D.setMaximumSize(QtCore.QSize(55, 16777215))
        self.npts_radial_2D.setInputMethodHints(QtCore.Qt.ImhDigitsOnly)
        self.npts_radial_2D.setObjectName("npts_radial_2D")
        self.horizontalLayout_8.addWidget(self.npts_radial_2D)
        self.npts_azim_2D = QtWidgets.QLineEdit(self.frame2D_header)
        self.npts_azim_2D.setMaximumSize(QtCore.QSize(55, 16777215))
        self.npts_azim_2D.setInputMethodHints(QtCore.Qt.ImhDigitsOnly)
        self.npts_azim_2D.setObjectName("npts_azim_2D")
        self.horizontalLayout_8.addWidget(self.npts_azim_2D)
        self.horizontalLayout_7.addLayout(self.horizontalLayout_8)
        self.frame2D_layout.addWidget(self.frame2D_header)
        self.frame2D_range = QtWidgets.QFrame(self.frame2D)
        self.frame2D_range.setMaximumSize(QtCore.QSize(16777215, 71))
        self.frame2D_range.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame2D_range.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2D_range.setObjectName("frame2D_range")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout(self.frame2D_range)
        self.horizontalLayout_9.setContentsMargins(5, 5, 5, 5)
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.gridLayout_2D = QtWidgets.QGridLayout()
        self.gridLayout_2D.setHorizontalSpacing(6)
        self.gridLayout_2D.setObjectName("gridLayout_2D")
        self.radial_autoRange_2D = QtWidgets.QCheckBox(self.frame2D_range)
        self.radial_autoRange_2D.setChecked(True)
        self.radial_autoRange_2D.setObjectName("radial_autoRange_2D")
        self.gridLayout_2D.addWidget(self.radial_autoRange_2D, 0, 4, 1, 1)
        self.unit_2D = QtWidgets.QComboBox(self.frame2D_range)
        self.unit_2D.setMaximumSize(QtCore.QSize(90, 16777215))
        self.unit_2D.setObjectName("unit_2D")
        self.unit_2D.addItem("")
        self.unit_2D.addItem("")
        self.gridLayout_2D.addWidget(self.unit_2D, 0, 0, 1, 1)
        self.label_to1_2 = QtWidgets.QLabel(self.frame2D_range)
        self.label_to1_2.setMaximumSize(QtCore.QSize(100, 30))
        self.label_to1_2.setObjectName("label_to1_2")
        self.gridLayout_2D.addWidget(self.label_to1_2, 1, 2, 1, 1)
        self.azim_low_2D = QtWidgets.QLineEdit(self.frame2D_range)
        self.azim_low_2D.setObjectName("azim_low_2D")
        self.gridLayout_2D.addWidget(self.azim_low_2D, 1, 1, 1, 1)
        self.radial_low_2D = QtWidgets.QLineEdit(self.frame2D_range)
        self.radial_low_2D.setObjectName("radial_low_2D")
        self.gridLayout_2D.addWidget(self.radial_low_2D, 0, 1, 1, 1)
        self.azim_autoRange_2D = QtWidgets.QCheckBox(self.frame2D_range)
        self.azim_autoRange_2D.setChecked(True)
        self.azim_autoRange_2D.setObjectName("azim_autoRange_2D")
        self.gridLayout_2D.addWidget(self.azim_autoRange_2D, 1, 4, 1, 1)
        self.azim_high_2D = QtWidgets.QLineEdit(self.frame2D_range)
        self.azim_high_2D.setObjectName("azim_high_2D")
        self.gridLayout_2D.addWidget(self.azim_high_2D, 1, 3, 1, 1)
        self.label_to2_2 = QtWidgets.QLabel(self.frame2D_range)
        self.label_to2_2.setObjectName("label_to2_2")
        self.gridLayout_2D.addWidget(self.label_to2_2, 0, 2, 1, 1)
        self.label_azim_2D = QtWidgets.QLabel(self.frame2D_range)
        self.label_azim_2D.setMinimumSize(QtCore.QSize(55, 0))
        self.label_azim_2D.setMaximumSize(QtCore.QSize(100, 50))
        self.label_azim_2D.setAlignment(QtCore.Qt.AlignCenter)
        self.label_azim_2D.setObjectName("label_azim_2D")
        self.gridLayout_2D.addWidget(self.label_azim_2D, 1, 0, 1, 1)
        self.radial_high_2D = QtWidgets.QLineEdit(self.frame2D_range)
        self.radial_high_2D.setObjectName("radial_high_2D")
        self.gridLayout_2D.addWidget(self.radial_high_2D, 0, 3, 1, 1)
        self.horizontalLayout_9.addLayout(self.gridLayout_2D)
        self.frame2D_layout.addWidget(self.frame2D_range)
        self.frame2D_buttons = QtWidgets.QFrame(self.frame2D)
        self.frame2D_buttons.setMaximumSize(QtCore.QSize(16777215, 71))
        self.frame2D_buttons.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame2D_buttons.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2D_buttons.setObjectName("frame2D_buttons")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout(self.frame2D_buttons)
        self.horizontalLayout_10.setContentsMargins(5, 5, 5, 32)
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setSpacing(6)
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.integrate2D = QtWidgets.QPushButton(self.frame2D_buttons)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.integrate2D.sizePolicy().hasHeightForWidth())
        self.integrate2D.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        self.integrate2D.setFont(font)
        self.integrate2D.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.integrate2D.setObjectName("integrate2D")
        self.horizontalLayout_11.addWidget(self.integrate2D)
        self.advanced2D = QtWidgets.QPushButton(self.frame2D_buttons)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced2D.sizePolicy().hasHeightForWidth())
        self.advanced2D.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        self.advanced2D.setFont(font)
        self.advanced2D.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.advanced2D.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.advanced2D.setObjectName("advanced2D")
        self.horizontalLayout_11.addWidget(self.advanced2D)
        self.horizontalLayout_10.addLayout(self.horizontalLayout_11)
        self.frame2D_layout.addWidget(self.frame2D_buttons)
        self.horizontalLayout_12.addLayout(self.frame2D_layout)
        self.verticalLayout.addWidget(self.frame2D)
        self.frame_3 = QtWidgets.QFrame(Form)
        self.frame_3.setMaximumSize(QtCore.QSize(16777215, 50))
        self.frame_3.setAutoFillBackground(True)
        self.frame_3.setStyleSheet("")
        self.frame_3.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout(self.frame_3)
        self.horizontalLayout_14.setContentsMargins(5, 0, 5, 0)
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_13.setSpacing(6)
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.pyfai_calib = QtWidgets.QPushButton(self.frame_3)
        self.pyfai_calib.setMinimumSize(QtCore.QSize(0, 20))
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(14)
        self.pyfai_calib.setFont(font)
        self.pyfai_calib.setObjectName("pyfai_calib")
        self.horizontalLayout_13.addWidget(self.pyfai_calib)
        self.get_mask = QtWidgets.QPushButton(self.frame_3)
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(14)
        self.get_mask.setFont(font)
        self.get_mask.setObjectName("get_mask")
        self.horizontalLayout_13.addWidget(self.get_mask)
        self.horizontalLayout_14.addLayout(self.horizontalLayout_13)
        self.verticalLayout.addWidget(self.frame_3)
        self.label_npts_1D.setBuddy(self.npts_1D)
        self.label_npts_2D.setBuddy(self.npts_radial_2D)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)
        Form.setTabOrder(self.axis1D, self.npts_1D)
        Form.setTabOrder(self.npts_1D, self.unit_1D)
        Form.setTabOrder(self.unit_1D, self.radial_low_1D)
        Form.setTabOrder(self.radial_low_1D, self.radial_high_1D)
        Form.setTabOrder(self.radial_high_1D, self.radial_autoRange_1D)
        Form.setTabOrder(self.radial_autoRange_1D, self.azim_low_1D)
        Form.setTabOrder(self.azim_low_1D, self.azim_high_1D)
        Form.setTabOrder(self.azim_high_1D, self.azim_autoRange_1D)
        Form.setTabOrder(self.azim_autoRange_1D, self.integrate1D)
        Form.setTabOrder(self.integrate1D, self.advanced1D)
        Form.setTabOrder(self.advanced1D, self.axis2D)
        Form.setTabOrder(self.axis2D, self.npts_radial_2D)
        Form.setTabOrder(self.npts_radial_2D, self.npts_azim_2D)
        Form.setTabOrder(self.npts_azim_2D, self.unit_2D)
        Form.setTabOrder(self.unit_2D, self.radial_low_2D)
        Form.setTabOrder(self.radial_low_2D, self.radial_high_2D)
        Form.setTabOrder(self.radial_high_2D, self.radial_autoRange_2D)
        Form.setTabOrder(self.radial_autoRange_2D, self.azim_low_2D)
        Form.setTabOrder(self.azim_low_2D, self.azim_high_2D)
        Form.setTabOrder(self.azim_high_2D, self.azim_autoRange_2D)
        Form.setTabOrder(self.azim_autoRange_2D, self.integrate2D)
        Form.setTabOrder(self.integrate2D, self.advanced2D)
        Form.setTabOrder(self.advanced2D, self.pyfai_calib)
        Form.setTabOrder(self.pyfai_calib, self.get_mask)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label1D.setText(_translate("Form", "1-D"))
        self.axis1D.setItemText(0, _translate("Form", "Radial"))
        self.label_npts_1D.setText(_translate("Form", "Points"))
        self.npts_1D.setText(_translate("Form", "3000"))
        self.npts_1D.setPlaceholderText(_translate("Form", "1000"))
        self.radial_autoRange_1D.setText(_translate("Form", "Auto"))
        self.unit_1D.setItemText(0, _translate("Form", "q (u\\u212Bu\\u207Bu\\u00B9)"))
        self.unit_1D.setItemText(1, _translate("Form", "2 u\\u03B8"))
        self.label_to1.setText(_translate("Form", "to"))
        self.azim_low_1D.setText(_translate("Form", "-180"))
        self.radial_low_1D.setText(_translate("Form", "0"))
        self.azim_autoRange_1D.setText(_translate("Form", "Auto"))
        self.azim_high_1D.setText(_translate("Form", "180"))
        self.label_to2.setText(_translate("Form", "to"))
        self.label_azim_1D.setText(_translate("Form", "Chi"))
        self.radial_high_1D.setText(_translate("Form", "5"))
        self.integrate1D.setText(_translate("Form", "Re-Integrate"))
        self.advanced1D.setText(_translate("Form", "Advanced..."))
        self.label2D.setText(_translate("Form", "2-D"))
        self.axis2D.setItemText(0, _translate("Form", "Q-Chi"))
        self.axis2D.setItemText(1, _translate("Form", "Qz-Qxy"))
        self.label_npts_2D.setText(_translate("Form", "Points"))
        self.npts_radial_2D.setText(_translate("Form", "1000"))
        self.npts_radial_2D.setPlaceholderText(_translate("Form", "1000"))
        self.npts_azim_2D.setText(_translate("Form", "1000"))
        self.radial_autoRange_2D.setText(_translate("Form", "Auto"))
        self.unit_2D.setItemText(0, _translate("Form", "q (u\\u212Bu\\u207Bu\\u00B9)"))
        self.unit_2D.setItemText(1, _translate("Form", "2 u\\u03B8"))
        self.label_to1_2.setText(_translate("Form", "to"))
        self.azim_low_2D.setText(_translate("Form", "-180"))
        self.radial_low_2D.setText(_translate("Form", "0"))
        self.azim_autoRange_2D.setText(_translate("Form", "Auto"))
        self.azim_high_2D.setText(_translate("Form", "180"))
        self.label_to2_2.setText(_translate("Form", "to"))
        self.label_azim_2D.setText(_translate("Form", "Chi"))
        self.radial_high_2D.setText(_translate("Form", "5"))
        self.integrate2D.setText(_translate("Form", "Re-Integrate"))
        self.advanced2D.setText(_translate("Form", "Advanced..."))
        self.pyfai_calib.setText(_translate("Form", "Calibrate"))
        self.get_mask.setText(_translate("Form", "Make Mask"))



# -*- coding: utf-8 -*-
"""
@author: thampy
"""

# Standard library imports
import requests
import time
import sys

# Other imports
import uuid
import pandas as pd
import re

try:
    from xdart.utils import find_between, find_between_r
except ImportError:
    sys.path.append('C:\\Users\\Public\\repos\\xdart')
    from xdart.utils import find_between, find_between_r

#  Hard coded server address - to be changed
bServer = "http://127.0.0.1:18085/SIS/"


def create_UUID():
    """Create unique ID to identify scan

    Returns:
        str -- unique ID
    """
    unique_uuid = uuid.uuid4().hex[:8].upper()
    return unique_uuid


def is_available(debug=False):
    """Check if SPEC is available to send commands.

    Keyword Arguments:
        debug {bool} --  Prints out SIS message if true (default: {False})

    Returns:
        bool -- True/False
    """
    t0 = time.time()
    for i in range(5):
        try:
            r = requests.get(bServer + "is_available")
            rv = r.json()['data']
            if debug: print(f'{i+1} Tries, {time.time()-t0:.3f} s')
            return rv
        except:
            pass
        
        time.sleep(0.05)
        
    if debug: print(f'{i+1} Tries, {time.time()-t0:0.3f} s')
    return False
          

def is_busy(debug=False):
    """Another way to check if SPEC is available to send commands.

    Keyword Arguments:
        debug {bool} --  Prints out SIS message if true (default: {False})

    Returns:
        bool -- True/False
    """
    t0 = time.time()
    for i in range(5):
        try:
            r = requests.get(bServer + "is_busy")
            rv = r.json()['data']
            if debug: print(f'{i+1} Tries, {time.time()-t0:.3f} s')
            return rv
        except:
            rv = True
        
        time.sleep(0.5)
        
    if debug: print(f'{i+1} Tries, {time.time()-t0:0.3f} s')
    return rv


def is_SpecBusy(debug=False):
    """Yet another way to heck if SPEC is busy.
    SpecInfoServer doesn't always track SPEC status well. 
    So it's good to try out different ways to make sure
 
    Keyword Arguments:
        debug {bool} --  Prints out SIS message if true (default: {False})

    Returns:
        bool -- True/False
    """
    t0 = time.time()
    
    rv = True
    
    #Read last line from SIS log
    last_line = get_sis_logs(num_entries=1)
    try:
        if ('reply' in last_line[0]) or ('Error' in last_line[0]):
            rv = False
    except:
        pass

    if debug: print(time.time() - t0)
    return rv

  
def wait_until_SPECfinished(debug=False, polling_time=1):
    """"Wait till SPEC finishes executing current command.
    

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})
        polling_time {float} -- Time to eait before querying again (default: {1})
    """
    while is_SpecBusy():
        time.sleep(polling_time)
        if not is_busy():
            break
        
    return
      
    
def wait_until_SPECfinished_old():
    """Deprecated"""
    t0 = time.time()
    while True:
        try:
            r = requests.get(bServer + "is_busy")
            is_busy = r.json()['data']
        except:
            is_busy = True
            
        if is_busy is False:
            break
        
        time.sleep(1)
    return time.time() - t0


def specCommand(cmd, queue=False, debug=False, print_console=True):
    """Function to send command to SPEC through SIS

    Arguments:
        cmd {str} -- SPEC command

    Keyword Arguments:
        queue {bool} -- Flag to enable queueing of command (default: {False})
        debug {bool} -- Flag to show output of SIS (default: {False})
        print_console {bool} -- Flag to print output to console (default: {True})
    """
    #if (not is_available()) or (status is 'running'):
    if (is_SpecBusy()) and (not is_available()):
        if queue:
            print('SPEC busy. Command Queued')
            wait_until_SPECfinished()
        else:
            print('SPEC not available')
            return
    
    r = requests.get(bServer + "get_remote_control")
    if debug: print(r.json())
        
    payload = {'spec_cmd': f"{cmd}"}
    if print_console:
        payload = {'spec_cmd': f"print '{cmd}';{cmd}"}
    r = requests.get(bServer + "execute_command", params=payload)
    if debug: 
        g = r.json()
        print(r, g['data'])
    
    print('Executed', '\n')
    #r = requests.get(bServer + "release_remote_control")
    return
    
    
def get_counter_mnemonics(debug=False):
    """Get the names of counters (Monitor, PD1 etc.)

    Keyword Arguments:
        debug {bool} -- Print output of SIS command (default: {False})

    Returns:
        [str] -- List of counter names (Monitor, PD1 etc.)
    """
    t0 = time.time()
    
    #Get the counters and put in dictionary
    r = requests.get(bServer + "get_all_counter_mnemonics", params={})
    counter_mnemonics = r.json()['data']
    
    if debug: print(time.time() - t0)
        
    return counter_mnemonics
        
    
def get_counter_values(debug=False):
    """Get the values of counters (Monitor, PD1 etc.)

    Keyword Arguments:
        debug {bool} -- Print output of SIS command (default: {False})

    Returns:
        [float] -- List of counter values
    """
    t0 = time.time()
    
    #Get the counters and put in dictionary
    r = requests.get(bServer + "get_all_counters", params={})
    counter_values = r.json()['data']
    
    if debug: print(time.time() - t0)
        
    return counter_values

    
def read_counters(debug=False):
    """Return dictionary containing counter names and values

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        dict -- counter names and values
    """
    t0 = time.time()
    
    #Get the counters and put in dictionary
    counter_mnemonics = get_counter_mnemonics()
    counter_values = get_counter_values()
    
    if debug: print(time.time() - t0)

    return dict(zip(counter_mnemonics, counter_values))


def get_motor_mnemonics(debug=False):
    """Get the names of motors

    Keyword Arguments:
        debug {bool} -- Print output of SIS command (default: {False})

    Returns:
        [str] -- List of motor names
    """
    t0 = time.time()
    
    #Get the motor positions and put in dictionary
    r = requests.get(bServer + "get_all_motor_mnemonics", params={})
    motor_mnemonics = r.json()['data']

    if debug: print(time.time() - t0)
    
    return motor_mnemonics


def get_motor_positions(debug=False):
    """Get the values of motors

    Keyword Arguments:
        debug {bool} -- Print output of SIS command (default: {False})

    Returns:
        [float] -- List of motor values
    """
    t0 = time.time()
    
    #Get the motor positions and put in dictionary
    r = requests.get(bServer + "get_all_motor_positions", params={})
    motor_positions = r.json()['data']

    if debug: print(time.time() - t0)
    
    return motor_positions


def read_motors(debug=False):
    """Return dictionary containing motor names and values

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        dict -- motor names and values
    """
    t0 = time.time()
    
    #Get the motor positions and put in dictionary
    motor_mnemonics = get_motor_mnemonics()
    motor_positions = get_motor_positions()
    
    if debug: print(time.time() - t0)
    
    return dict(zip(motor_mnemonics, motor_positions))


def get_current_scan_details(debug=False):
    """Get details of current scan

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        str -- scan status (busy, done, unknown)
    """
    t0 = time.time()
    
    #Get current scan details
    for i in range(10):
        try:
            r = requests.get(bServer + "get_current_scan_details", params={})
            if debug: print(time.time() - t0)
            return r.json()['data']
        except:
            pass
        
    return 'unknown'
    
    
def get_scan_status(debug=False):
    """Get details of current scan

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        str -- scan status (busy, done, unknown)
    """
    #Get current scan details
    try: 
        r = requests.get(bServer + "get_current_scan_details", params={})
        scan_details = r.json()['data']
        return re.search('status=(.*) ID', scan_details).group(1)
    except:
        return 'unknown'
    
    
def get_last_scan_details(debug=False):
    """Get details of last finished scan

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        dict -- scan details (scan_type, scan_number, scan_sfx, scan_pts)
    """
    t0 = time.time()
    
    #Get current scan details
    while get_scan_status() != 'finished':
        time.sleep(0.1)
        
    try: 
        r = requests.get(bServer + "get_current_scan_details", params={})
        scan_details = r.json()['data']
        
        scan_type = re.search('X_lbl=(.*) Y_lbl', scan_details).group(1)
        scan_npts = re.search('pt=(.*) type', scan_details).group(1)[-3:]

        scan_name = re.search('ID=(.*) pt', scan_details).group(1)
        scan_number = find_between_r(scan_name, '_', '')
        scan_sfx = scan_name[:-len(scan_number)]

        return {'scan_type': scan_type,
                'scan_number': scan_number, 
                'scan_sfx': scan_sfx, 
                'scan_npts': scan_npts}
    except:
        return 'Cannot determine scan details'    
    

def get_user():
    """Return name of current SPEC user

    Returns:
        str -- SPEC user name
    """
    t0 = time.time()
    
    #Get user name
    r = requests.get(bServer + "get_user", params={})
    user = r.json()['data']
    
    return user


def abort_command():
    """Abort current command

    Returns:
        str -- Result of abort command returned by SIS
    """
    #Abort - similar to Ctrl-C in SPEC
    r = requests.get(bServer + "abort_command", params={})
    result = r.json()['data']
    
    return result


def set_sis_logging(set_logging_on=True, debug=False):
    """Enable SIS logging that is useful for debugging

    Keyword Arguments:
        set_logging_on {bool} -- Enables logging (default: {True})
        debug {bool} -- Print output of SIS command to enable logging (default: {False})

    Returns:
        str -- result of SIS command to enable logging (success/fail)
    """
    t0 = time.time()
    
    #Get current scan details
    payload = {'set_logging_on': set_logging_on}
    r = requests.get(bServer + "get_sis_logs", params=payload)
    result = r.json()['data']
    
    if debug: print(r, time.time() - t0)
    return result


def get_sis_logs(num_entries=None, debug=False):
    """Get last num_entries lines of SIS log. Returns the whole log if num_entries is None

    Keyword Arguments:
        num_entries {int} -- Last number of lines of log to return (default: {None})
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        [str] -- Lines of SIS log
    """
    t0 = time.time()
    
    #Get log from SIS log
    payload = {'num_entries': num_entries}
    r = requests.get(bServer + "get_sis_logs", params=payload)
    log = r.json()['data']
    
    if debug: print(time.time() - t0)
    return log


def get_plot_points(num_pts=None, debug=False, columns=None):
    """Return all the values recorded for scan points. The last num_pts are
    reported. If num_pts is None, the entire last scan is reported 

    Keyword Arguments:
        num_pts {int} -- Number of scan points to report (default: {None})
        debug {bool} -- Print output of SIS command if True (default: {False})
        columns {[str]} -- list of column values to return (default: {None})

    Returns:
        pandas dataframe -- dataframe containing columns and related values
    """
    t0 = time.time()
    
    #Get log from SIS log
    payload = {'num_pts': num_pts}
    r = requests.get(bServer + "get_plot_points", params=payload)
    data = pd.DataFrame(r.json()['data'], columns=columns)
    
    if debug: print(time.time() - t0)
    
    return data


def get_spec_result(debug=False):
    """Return result of last SPEC command

    Keyword Arguments:
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        str -- Result of last scan command
    """
    t0 = time.time()
    
    #Get log from SIS log
    payload = {}
    r = requests.get(bServer + "retrieve_result", params=payload)
    log = r.json()['data']
    
    if debug: print(time.time() - t0)
    return log


def get_console_output(idx=1, debug=False):
    """Return the console output for idx number of lines

    Keyword Arguments:
        idx {int} -- Number of lines of console output to return (default: {1})
        debug {bool} -- Print output of SIS command if true (default: {False})

    Returns:
        [str] -- Console output
    """
    t0 = time.time()
    
    #Get log from SIS log
    payload = {'N': idx}
    r = requests.get(bServer + "get_console_output_buffer", params=payload)
    log = r.json()['data']
    
    if debug: print(time.time() - t0)
    return log


import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio

sys.path.append('C:\\Users\\Public\\repos\\xdart')
from xdart.modules.pySSRL_bServer.bServer.BL_Error import *
from xdart.modules.pySSRL_bServer.bServer.useful_func import *

class BL_Communication():
    """Different communication protocals can be added and extended through this class. A few base functions like writeCommand and readResponse should not depend on the communication protocal and will be implemented here."""
    async def writeCommand(self, mCmd, terminator=None):
        """ Encode the string and add the appropiate send terminator to the command.
        """
        debug("writeCommand: '{}'".format(mCmd))
        if terminator is None:
            mCmd += self.SEND_TERMINATOR

        encCmd = mCmd.encode(self.STR_ENCODING)


        await self.write(encCmd)
        debug("writeCommand: finished writing command")

    async def readResponse(self, terminator=None, listSeparator=None):
        """Read the response from the beamline, strip the terminators, and break the response into a list seperated by 'listSeparator'.
        """

        if terminator is None:
            terminator = self.RECEIVE_TERMINATOR

        if listSeparator is None:
            listSeparator = self.RECEIVE_TERMINATOR

        try:
            #print("Before read")
            bytesReceived = await (self.read(terminator))

            print("after read: '{}'".format(bytesReceived))

        except self.CommandError:
            warning("got command error in read")
            raise
        except:
            warning("Unexpected error in read:", sys.exc_info()[0])
            raise

        returnedText = bytesReceived.decode(self.STR_ENCODING)

        strippedText = returnedText.rstrip(terminator)
        parsedList = strippedText.split(listSeparator)

        filteredList = list(filter(None, parsedList)) #Remove the empty items from the list
        return filteredList

    
    #### These functions should be overloaded by the subclass for each communication protocol
    #def read(self, terminator=None):
    #    pass

    #def write(self, toSend):
    #    pass
       
    #def __init__(self, mi):
    #    pass


class tmp_bi():
    """TEMP class to let me assign a few things the BL_Interaction class will handle later."""
    def __init__(self, loop, beamline_name):
        self.loop = loop
        self.beamline_name = beamline_name
    
    
class BL_SPEC(BL_Communication):
    """Functions for communicating with SPEC beamlines through TCPIP sockets. This is the fairly low level stuff.
    One could imagine writing similar classes for ICS """
    def __init__(self, bi):
        """Set the IP address and port for the beamline by loading from the config file. Configure buffer sizes, encodings, and terminators. Call connect. Note: 'mi' is the motor interactor that will hold this object."""

        #Record a reference to the motor interactor that holds this object
        self.bi = bi

        #Init dict for asyncio socket
        self.s = {'reader': None, 'writer': None}

        #Initialize communication paramters that don't belong in the buffer
        self.BUFFER_SIZE = 1024*1024
        self.TIMEOUT = 1 #This is in seconds
        self.STR_ENCODING = 'ascii'
        self.SEND_TERMINATOR = "\n"
        self.RECEIVE_TERMINATOR = "\n"
        self.MAX_READ_TIME = 5
        #Load up the config file for this beamline
#        self.load_beamline_config(self.bi.beamline_name)

        #This is a bit crude. Lets set the config filename/path here.
        self.beamline_config_filename = './config/beamline_catalogue.yml'

        parsed_config = load_yaml_config(self.beamline_config_filename)
        try:

            #Take the elements in the config and add them to the object
            bl_config = parsed_config[self.bi.beamline_name]

            for (key, val) in bl_config.items():
                print("adding '{}' -> '{}'".format(key, val))
                setattr(self, key, val)
        except:
            print('BL_SPEC: init: hit an error: {}', sys.exc_info()[0])
            raise


        #Get the async components setup
        asyncio.ensure_future(self.setup(), loop=self.bi.loop)

    async def setup(self):
        # Parse the config file and add all of the relevent attributes to the LMD_TCPIP object
        debug("setup: doing async components of beamline setup")
        # Connect to the motor
        try:
            await self.connect()
        except:
            raise RuntimeError("setup: problem in setup of {}".format(self.bi.beamline_name))
        else:
            print("motor '{}' setup successful. socket object = {}".format(self.bi.beamline_name, self.s))


    async def reconnect(self):
        """CLose and then open the socket again"""
        try:
            self.close()
            await self.connect()
        except SPECcommunicationError:
            debug("reconnect(): Problem communicating w/ SPEC during reconnect")
        except:
            debug("reconnect(): Some other exception")
            raise
        else:
            info("reconnect(): reconnected!")
            print("Reconnect worked!")


    async def connect(self):
        """Open the socket and connect to the motor"""
        tcpip = self.TCP_IP
        tcpport = self.TCP_PORT

        info("connecting to {}:{}".format(tcpip, tcpport))
        self.connected = False
        
        try:
            self.s['reader'], self.s['writer'] = await asyncio.wait_for(asyncio.open_connection(host=tcpip, port=tcpport, loop=self.bi.loop), timeout=self.TIMEOUT, loop=self.bi.loop)
            self.s['writer'].transport.set_write_buffer_limits(low=0, high=0)
        except asyncio.TimeoutError as e:
            warning("Socket Error: Problem connecting to {}:{}".format(tcpip, tcpport))
            raise SPECCommunicationError(expression="connect()", message="Hit an asynio timeout error")
        except:
            warning("connect: BIG Problem connecting to {}:{}".format(tcpip, tcpport))
            warning("connect: Unexpected error in read:", sys.exc_info()[0])
            print("hit exception in connect")
            raise
        else:
            self.connected = True

            info("connect(): done connecting...")
            debug("connect(): {}".format(self.s))
            info("connect(): Socket connected: {}".format(self.connected))


    async def read(self, terminator=None):
        debug("read(): starting")
        if terminator is None:
            terminator = self.RECEIVE_TERMINATOR

        print("terminator:", terminator)
        readBytes = b""
        readStart = time.time()
        try:
            while True:
                #Read the buffer. This is blocking for now
                cBuf = await asyncio.wait_for(self.s['reader'].read(self.BUFFER_SIZE), timeout=self.TIMEOUT, loop=self.bi.loop)
                readBytes += cBuf

                #Test to see if we have made it to the end of the response by looking for the terminator
                if readBytes[(-len(terminator)):] == terminator.encode():
                    debug("Read: '{}'".format(readBytes.decode(self.STR_ENCODING)))
                    break


                #If something goes wrong, we might not get the terminitor that we are expecting.
                if (time.time()-readStart) > self.MAX_READ_TIME:
                    warning("read: terminator not found. Read '{}'".format(readBytes))
                    raise SPECCommandError('read', "Read {} bytes. Terminator not found. data: '{}'".format(len(readBytes), readBytes))

        except asyncio.TimeoutError:
            debug("read: timeout error: read '{}'".format(readBytes))
            raise SPECCommunicationError(expression="read()", message="asyncio timed out")
            raise

        except:
            warning("read(): unexpected error in reading socket:", sys.exc_info()[0])
            raise

        finally:
            debug("read(): returning '{}'".format(readBytes))
            return readBytes

    async def write(self, toSend):
        """Send data to the socket. """
        debug("write(): '{}'".format(toSend))
        try:
            #Write to the streamwriter
            write_result = self.s['writer'].write(toSend) #Returns none on success

            #drain the writer. This should in theory block if the transport hasn't happened yet but I can't get that to work. Right now the wait_for doesn't appear to do anything as the drain routine is not returning a future..
            drain_result = await asyncio.wait_for(self.s['writer'].drain(), timeout=self.TIMEOUT, loop=self.bi.loop)

            debug("write(): drain future: {}".format(drain_result))

        except asyncio.TimeoutError:
            raise SPECCommunicationError(expression="write()", message="asyncio timed out")
        except:
            critical("write: hmmmm.... critical error in write: ",  sys.exc_info()[0])
            raise
            
    def close(self):
        """Just close the socket and set the connected property to false."""
        try:
            debug("close(): closing socket")
            self.s['writer'].close() #Close the writer. this should close the connection
        except:
            warning("close: Error closing socket")
            raise
        else:
            debug("close(): close successful")
            self.connected = False

    def __del__(self):
        """Cleanup the beamline object. Basically, just make sure the socket connection is closed."""
        # debug("called __del__ on beamline interactor")
        self.close()




import asyncio


class OracleFrame(FrameObject):

    def __init__(self, frame_pool, var_list):
        """Create the oracle frame to follow motor interactors that cover all the variables in var_list. Do not
        actually attach the followers here since that will make it awkward to return the futures each var will
        actually track"""
        #Create the mi_list based off the unique mi values in var_list
        self.mi_list = self.get_motor_interactors_from_var_list(var_list)

        #print("before common init")
        #run the common init routine to create and register futures and generate the UUID
        (self._common_init(frame_pool, self.mi_list))

        #Finally, set us to be attachable for followers. I probably could have done this earlier.
        #self.lock_and_set_attachable(True)
        self.allow_attachement = True

        #self.attach_followers(var_list=var_list)


        #add frame to pulpits required for var_list. Reserves our place in the queue
        self.attach_pulpits()

        #Start the async oracle meaurement
 #       self.start()

    def start(self):
        """Launches the asyncronous routine that actually updates the motor position. """
        #Start the oracle frame async routine.
        self.running_future = asyncio.ensure_future(self.run_oracle_frame(), loop=self.frame_pool.loop)




    def get_motor_interactors_from_var_list(self, var_list):
        """get all of the unique motor interactors from a list of variables."""

        #Pretty simple trick. Get all of the motor_interactors in a list, convert it to a set (which removes duplicates) and then convert that back to a list.
        mi_list = list(set([motor_var.mi for motor_var in var_list]))

        return mi_list


    def allows_followers(self, required_var_list):
        """Check if this frame can follow all of the required_vars. Oracle frames always allow followers if they are in the motor interactor associated with this frame."""
        try:
            for var in required_var_list:
                #Check that the current variable has a motor interactor in the oracle frames list of motor interactors
                self.mi_list.index(var.mi)
        except ValueError:
            #This means the variable was associated with a motor interactor not in the oracle frames MI list
            return False

        else:
            #All of the required_vars are within the scope of this oracle frame
            return True





    async def run_oracle_frame(self):
        """Lets implement the actual oracle frame. It runs asyncronously and is launched by the http handler. First, we
        wait on the start futures. Second, we turn off the ability to attach. Thrid, we loop through all of the motor
        interactors and execute updates for any routines that have followers. Finally, we set the end futures and end
        the function."""


        #Add a context manager to handle errors


        #wait on the start futures to be set by the pulpits



        print("Starting run_oracle_frame", self)
        for (pulpit_obj, future) in self.start_future.items():
            print("run_oracle_frame checking future. Future: {}  Pulpit: {}".format(future, pulpit_obj))
            if future.done() is False:
                print('  waiting on Future')
                await future
            else:
                print('  continuing')

        t0 = time.time()
        print("stopping attachement")
        #Stop attachement
        self.allow_attachement = False

        print("  mi_list: ", self.mi_list)
        for mi in self.mi_list:

            var_dict = mi.motor_variables.get_motor_variables()

            #print(var_dict)

            for (var_name, motor_var) in var_dict.items(): #This is a little clunky but will work

                #print("Looking at {}: {}".format(var_name, motor_var))
                #Technically, we could just run the update routine... but lets check so we can be sure what we are updating...

                num_followers = motor_var.get_followers(self.frame_uuid)
                #print(" num_followers: ", num_followers)
                if num_followers is not None:
                    print("Updating followers of variable '{}'".format(var_name))
                    await motor_var.read_value(self.frame_uuid)
                    print("  updating {}: {}s".format(motor_var.name, time.time()-t0))
                    #motor_var.update_followers(oracle_frame.frame_uuid)



        #print("setting end futures")
        #Set the end routines
        for (pulpit_obj, future) in self.end_future.items():
            print("run_oracle_frame setting end. Future: {}  Pulpit: {}".format(future, pulpit_obj))
            future.set_result("  finished {}".format(time.time()))
            print("  end future set ", future)

        self.flag_finished = True



import sys
import time
import logging

from watchdog.observers import Observer  
from watchdog.events import PatternMatchingEventHandler

class MyHandler(PatternMatchingEventHandler):
    patterns = ["*.pdi"]

    def process(self, event):
        """
        event.event_type 
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        print(event.src_path, event.event_type)  # print now only for degug

    def on_modified(self, event):
        self.process(event)

    def on_created(self, event):
        self.process(event)
        
        
if __name__ == '__main__':
    args = sys.argv[1:]
    observer = Observer()
    observer.schedule(MyHandler(), path=args[0] if args else '.', recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()


import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio
from aiohttp import web
import functools
from time import gmtime, strftime

sys.path.append('C:\\Users\\Public\\repos\\xdart')
from xdart.modules.pySSRL_bServer.bServer.bl_communication import *
from xdart.modules.pySSRL_bServer.bServer.bl_command import *
from xdart.modules.pySSRL_bServer.bServer.BL_Error import *
from xdart.modules.pySSRL_bServer.bServer.bl_interaction import *


async def handleGET_motor_position(request):
    """Quick routine to test reading the motor position from SPEC"""
    t0 = time.time()

    motor_name = request.match_info.get('motor_name', None)
    print("getting motor variable '{}' ".format(motor_name))

    bi = request.app['bi']

    comm = bi.beamline

    cmd_text = "?mp {}".format(motor_name)

    cmd = BLCommand(comm, cmd_text, needsResponse=True)

    await cmd.execute()

    print(cmd)

    response = cmd.typecastResponse([float])
    print("Response: '{}'".format(response))
    print("\n\n\nRequest took: {}s\n\n".format(time.time() - t0))

    #convertedResults = {'hi': 'there', 'position': response}
    convertedResults = {'position': response}
    return web.json_response(convertedResults)


async def handleGET_tt(request):
    """Quick routine to test reading the motor position from SPEC"""
    t0 = time.time()
    print("starting tt")
    bi = request.app['bi']

    comm = bi.beamline

    cmd1 = BLCommand(comm, "!rqc", needsResponse=True)
    cmd2 = BLCommand(comm, "!cmd umv rail3_y -5", needsResponse=True)
    cmd3 = BLCommand(comm, "!rlc", needsResponse=True)

    cmd = cmd1 + cmd2 + cmd3
    await cmd.execute()

    #response = cmd.response
    print(cmd)

    print("Response: '{}'".format(cmd))
    print("\n\n\nRequest took: {}s\n\n".format(time.time() - t0))

    convertedResults = {'hi': 'there', 'position': cmd}
    return web.json_response(convertedResults)


async def handleGET_test(request):
    """Quick routine to test reading the motor position from SPEC"""
    t0 = time.time()
    print("starting test")
    bi = request.app['bi']

    try:

        await bi.sis.get_remote_control()
        response = await bi.sis.are_we_in_control()
#        starting_at = await bi.sis.get_console_output_buffer(get_buffer_index=True)
#        print("Starting index: {}".format(starting_at))
#        await bi.sis.execute_command("ct 0.3")

#        #await asyncio.sleep(2)
#        response = await bi.sis.get_console_output_buffer(N=starting_at, return_after=True)
#        print("What happened:\n\n", response)
#        answer = await bi.sis.retrieve_result()
#        print("Answer:\n", answer)

        await bi.sis.release_remote_control()

    except:
        print("problem in test: {}".format(sys.exc_info()[0]))
        raise

    print("Response: '{}'".format(response))
    print("\n\n\nRequest took: {}s\n\n".format(time.time() - t0))

    convertedResults = {'hi': 'there', 'data': response}
    return web.json_response(convertedResults)


async def handleGET_directSISinteraction(request):
    """Quick routine to test reading the motor position from SPEC"""
    t0 = time.time()
    print("starting direct SPEC Infoserver interaction")
    bi = request.app['bi']


    command_name = request.match_info.get('command_name', None)
    command_params = {}
    #Get the parameters passed as part of the GET request
    query = request.query
    print("Found arguments: (Key-> Val)\n")

    for (key, val) in query.items():
        print("   '{}'->'{}'".format(key, val))
        if val == "True" or val == "true":
            command_params[key] = True
        elif val == "False" or val == "false":
            command_params[key] = False
        elif val == "None" or val == "none":
            command_params[key] = None
        else:
            command_params[key] = val

    response = {}
    try:

        sis_command = getattr(bi.sis, command_name, None)
        print("sis_command object: ", sis_command)

        response['help'] = sis_command.__doc__
        response['data'] = await sis_command(**command_params)
        pass
    except:
        print("problem in test: {}".format(sys.exc_info()[0]))
        response = {'error': sys.exc_info()[0]}

#        raise

    print("Response: '{}'".format(response))
    print("\n\n\nRequest took: {}s\n\n".format(time.time() - t0))

    return web.json_response(response)

########################################################################################################################
########################################################################################################################


log_datetime = strftime("%Y-%m-%d_%H_%M_%S", gmtime())
logging.basicConfig(filename='bServerLog__{}.log'.format(log_datetime), level=logging.DEBUG)

#Setup the event loop
loop = asyncio.new_event_loop()
loop.set_debug(True)
asyncio.set_event_loop(loop)


#Initialize the connection to SPEC
temp_bi = BL_Interaction(loop=loop, beamline_name='bl11-3_testConfig')

#Initialize the http web application and register routes
app = web.Application()

app.router.add_get("/motor/{motor_name}/position", handleGET_motor_position)
app.router.add_get("/tt", handleGET_tt)
app.router.add_get("/test", handleGET_test)
app.router.add_get("/SIS/{command_name}", handleGET_directSISinteraction)


#app.router.add_post("/motor/{motor_name}", handlePOST_move_motor)

app['bi'] = temp_bi

web.run_app(app, host='127.0.0.1', port=18085)#, loop=loop)



from mServer.frame_objects import *

from mServer.OracleFrame import *
from mServer.ControlFrame import *
from mServer.PulpitObject import *

import asyncio
import uuid
import time


class FramePool():
    """Create and hold references to all active Frame objects. While control frames are always for a specific action,
    the creation of an oracle frame is only required when nothing is available in the current pool. Thus, it is
    managed here"""

    def __init__(self, loop, mi_list):

        # This is a lock used to prevent any funny business when looking through the list of avialable frames
        # for one that can attach followers. I'm not actually sure this is necessary while we use asyncio
        # (Interestingly asyncio.Lock is not threadsafe - which is OK since we are using asyncio's framework)
        self.attachable_lock = asyncio.Lock()

        #This holds all of the active lists. Make sure to lock when accessing. Be careful not to deadlock!
        self.frame_list = []

        #hold a list of all the motor interactors accessible by this framepool
        self.mi_list = mi_list

        #Hold onto the event loop in which to register everything.
        self.loop = loop

        info('FramePool initialized')


    def find_attachable_frame(self, follower_vars):
        """Search the FramePool for anyone queued up that could provide the information for the followers. If no
        follower exists, create an oracle frame for all motor interactors specified by the follower_vars. Note: This
        will be a linked oracle frame if multiple mi's exist. The frame is returned."""

        print('starting cleanup')
        # Start with some house cleaning.
        self.cleanup_frame_list()

        print('done with cleanup')
        # Make sure we have exclusive access to the attachable list
        #       with (yield from self.attachable_lock):

        # Create an iterator that looks for frames in the frame_list that allow all of the followers
        filtered_frames = filter(lambda x: x.flag_finished is False and x.allows_followers(follower_vars) is True and x.allow_attachement is True, self.frame_list)

        print(filtered_frames)
        # Try to attach to the frame
        try:
            attachable_frame = next(filtered_frames)
            # return the frame_uuid back to the caller
        except StopIteration:  # No FrameObject found that will give the follower the necessary information.
            # So we were not sucessful in attaching followers. Lets Create an oracle frame
            print('creating oracle frame')
            oracle_frame = OracleFrame(frame_pool=self, var_list=follower_vars)
            print(' done creating oracle frame')
            return oracle_frame
        else:  # Alright, we have an attachable_frame. attach followers and return
            attachable_frame.attach_followers(follower_vars)
            return attachable_frame

    def cleanup_frame_list(self):
        """Remove all frames from the frame_list that have flag_delete set to True. frame_list is locked."""
        #This is a bit crude right now and it does require that I actively set FrameObjects to be deleted
#        with (yield from self.attachable_lock):
        self.frame_list = [ frame for frame in self.frame_list if frame.flag_delete is False ]


    def generate_frame_uuid(self):
        """generate unique frame IDs. This is realtively easy to do since all ID's for all motors are generated
         by a s ingle instance of FramePool. This is an 8-digit hexadecimal number which gives us 4,294,967,296 values before
         the sequence could repeat. Since each command takes ~10ms at least, it would take roughly 500 days for something to happen....
         Actually, nothing would happen because commands that old would be taken care of by the garbage collector long long long before
         that happened. """

        frame_uuid = uuid.uuid4().hex[:8].upper()
        return frame_uuid


    def get_attachable_variables(self, mi_list=None, include_hidden_variables = True):
        """Return a list of variables that can be set / get (followed). The return object is a dictionary with the
        motor interactor as the key with a dictionary of motor variaibles. include_hidden_variables is true by default"""
        ret_dict = {}

        if mi_list is None:
            mi_list = self.mi_list

        for mi in mi_list:
            ret_dict[mi] = mi.motor_variables.get_motor_variables(include_hidden_variables=include_hidden_variables)

        return ret_dict



    async def setup_background_tasks(self):
        """Start the asynchronous background tasks that concern the frame pool"""
        info('FramePool background tasks started')

        #Start frame_list cleanup routine
        asyncio.ensure_future(self.cleanup_old_frames(), loop=self.loop)

    async def cleanup_old_frames(self, cleanup_interal = 5, remove_frames_older_than = 300):
        """Run garbage collection in the background while """
        while True:
            #Wait for a little bit
            await asyncio.sleep(cleanup_interal, loop=self.loop)

            start_time = time.time()
            #Delete elements flagged for
            self.frame_list = [frame for frame in self.frame_list if frame.flag_delete is True]

            #Delete elements that are too old
            current_time = time.time()
            self.frame_list = [frame for frame in self.frame_list if frame.flag_finished is True and (current_time - frame.time_created) > remove_frames_older_than]

            print("cleanup_old_frames: it took {} seconds to cleanup".format(time.time() - start_time))





from threading import Condition, _PyRLock
import os
from pathlib import Path
import pandas as pd
import numpy as np
from pyFAI.multi_geometry import MultiGeometry

from .arch import EwaldArch
from .arch_series import ArchSeries
from xdart.utils.containers import int_1d_data, int_2d_data
from xdart.utils.containers import int_1d_data_static, int_2d_data_static
from xdart import utils

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)


class EwaldSphere:
    """Class for storing multiple arch objects, and stores a MultiGeometry
    integrator from pyFAI.

    Attributes:
        arches: ArchSeries, list of arches indexed by their idx value
        bai_1d: int_1d_data object, stores result of 1d integration
        bai_1d_args: dict, arguments for invidivual arch integrate1d
            method
        bai_2d: int_2d_data object, stores result of 2d integration
        bai_2d_args: dict, arguments for invidivual arch integrate2d
            method
        data_file: str, file to save data to
        file_lock: lock for ensuring one writer to hdf5 file
        mg_args: arguments for MultiGeometry constructor
        mgi_1d: int_1d_data object, stores result from multigeometry
            integrate1d method
        mgi_2d: int_2d_data object, stores result from multigeometry
            integrate2d method
        multi_geo: MultiGeometry instance
        name: str, name of the sphere
        scan_data: DataFrame, stores all scan metadata
        sphere_lock: lock for modifying data in sphere

    Methods:
        add_arch: adds new arch and optionally updates other data
        by_arch_integrate_1d: Runs 1 dimensional integration of each
            arch individually and sums the result, stored in bai_1d
        by_arch_integrate_2d: Runs 2 dimensional integration of each
            arch individually and sums the result, stored in bai_2d
        load_from_h5: loads data from hdf5 file
        set_multi_geo: sets the MultiGeometry instance
        multigeometry_integrate_1d: wrapper for MultiGeometry
            integrate1d method, result stored in mgi_1d
        multigeometry_integrate_2d: wrapper for MultiGeometry
            integrate2d method, result stored in mgi_2d
        save_bai_1d: Saves only bai_1d to the data_file
        save_bai_2d: Saves only bai_2d to the data_file
        save_to_h5: saves data to hdf5 file
        set_multi_geo: instatiates the multigeometry object, or
            overrides it if it already exists.
    """

    def __init__(self, name='scan0', arches=[], data_file=None,
                 scan_data=pd.DataFrame(), mg_args={'wavelength': 1e-10},
                 bai_1d_args={}, bai_2d_args={},
                 static=False, gi=False, th_mtr='th', series_average=False,
                 overall_raw=0, single_img=False,
                 global_mask=None, poni_dict={}
                 ):
        """name: string, name of sphere object.
        arches: list of EwaldArch object, data to intialize with
        data_file: str, path to hdf5 file where data is stored
        scan_data: DataFrame, scan metadata
        mg_args: dict, arguments for Multigeometry. Must include at
            least 'wavelength' attribute in Angstroems
        bai_1d_args: dict, arguments for the integrate1d method of pyFAI
            AzimuthalIntegrator
        bai_2d_args: dict, arguments for the integrate2d method of pyFAI
            AzimuthalIntegrator
        """
        super().__init__()
        self.file_lock = Condition()
        if name is None:
            self.name = os.path.split(data_file)[-1].split('.')[0]
        else:
            self.name = name
        if data_file is None:
            self.data_file = name + ".hdf5"
        else:
            self.data_file = data_file

        self.static = static
        self.gi = gi
        self.th_mtr = th_mtr
        self.single_img = single_img
        self.series_average = series_average

        if arches:
            self.arches = ArchSeries(self.data_file, self.file_lock, arches,
                                     static=self.static, gi=self.gi)
        else:
            self.arches = ArchSeries(self.data_file, self.file_lock,
                                     static=self.static, gi=self.gi)
        self.scan_data = scan_data

        self.mg_args = mg_args
        if len(arches) > 0:
            self.multi_geo = MultiGeometry([a.integrator for a in arches], **mg_args)

        self.bai_1d_args = bai_1d_args
        self.bai_2d_args = bai_2d_args
        self.mgi_1d = int_1d_data()
        self.mgi_2d = int_2d_data()
        self.sphere_lock = Condition(_PyRLock())

        if self.static:
            self.bai_1d = int_1d_data_static()
            self.bai_2d = int_2d_data_static()
        else:
            self.bai_1d = int_1d_data()
            self.bai_2d = int_2d_data()

        self.overall_raw = overall_raw
        self.global_mask = global_mask
        self.poni_dict = poni_dict

    def reset(self):
        """Resets all held data objects to blank state, called when all
        new data is going to be loaded or when a sphere needs to be
        purged of old data.
        """
        with self.sphere_lock:
            self.scan_data = pd.DataFrame()
            self.mgi_1d = int_1d_data()
            self.mgi_2d = int_2d_data()
            self.arches = ArchSeries(self.data_file, self.file_lock,
                                     static=self.static, gi=self.gi)
            self.global_mask = None
            if self.static:
                self.bai_1d = int_1d_data_static()
                self.bai_2d = int_2d_data_static()
            else:
                self.bai_1d = int_1d_data()
                self.bai_2d = int_2d_data()
            self.overall_raw = 0

    def add_arch(self, arch=None, calculate=True, update=True, get_sd=True,
                 set_mg=True, **kwargs):
        """Adds new arch to sphere.

        args:
            arch: EwaldArch instance, arch to be added. Recommended to
                always pass a copy of an arch with the arch.copy method
                or intialize with kwargs
            calculate: whether to run the arch's calculate methods after
                adding
            update: bool, if True updates the bai_1d and bai_2d
                attributes
            get_sd: bool, if True tries to get scan data from arch
            set_mg: bool, if True sets the MultiGeometry attribute.
                Takes a long time, especially with longer lists.
                Recommended to run set_multi_geo method after all arches
                are loaded.
            kwargs: If arch is None, used to intialize the EwaldArch,
                see EwaldArch for arguments.

        returns None
        """
        with self.sphere_lock:
            if arch is None:
                arch = EwaldArch(**kwargs)
            if calculate:
                arch.integrate_1d(global_mask=self.global_mask, **self.bai_1d_args)
                arch.integrate_2d(global_mask=self.global_mask, **self.bai_2d_args)
            arch.file_lock = self.file_lock
            self.arches = self.arches.append(pd.Series(arch, index=[arch.idx]))
            self.arches.sort_index(inplace=True)

            if arch.scan_info and get_sd:
                ser = pd.Series(arch.scan_info, dtype='float64')
                if list(self.scan_data.columns):
                    try:
                        self.scan_data.loc[arch.idx] = ser
                    except ValueError:
                        print('Mismatched columns')
                else:
                    self.scan_data = pd.DataFrame(
                        arch.scan_info, index=[arch.idx], dtype='float64'
                    )
                self.scan_data.sort_index(inplace=True)
                with self.file_lock:
                    with utils.catch_h5py_file(self.data_file, 'a') as file:
                        compression = 'lzf'
                        if self.static:
                            compression = None
                        utils.dataframe_to_h5(self.scan_data, file,
                                              'scan_data', compression)
            if update:
                self._update_bai_1d(arch)
                self._update_bai_2d(arch)
            if set_mg:
                self.multi_geo = MultiGeometry(
                    [a.integrator for a in self.arches], **self.mg_args
                )

            self.overall_raw += (arch.map_raw - arch.bg_raw)

    def by_arch_integrate_1d(self, **args):
        """Integrates all arches individually, then sums the results for
        the overall integration result.

        args: see EwaldArch.integrate_1d. If any args are passed, the
            bai_1d_args dictionary is also updated with the new args.
            If no args are passed, uses bai_1d_args attribute.
        """
        if not args:
            args = self.bai_1d_args
        else:
            self.bai_1d_args = args.copy()
        with self.sphere_lock:
            if self.static:
                self.bai_1d = int_1d_data_static()
            else:
                self.bai_1d = int_1d_data()

            for arch in self.arches:
                arch.integrate_1d(global_mask=self.global_mask, **args)
                self.arches[arch.idx] = arch
                self._update_bai_1d(arch)

    def by_arch_integrate_2d(self, **args):
        """Integrates all arches individually, then sums the results for
        the overall integration result.

        args: see EwaldArch.integrate_2d. If any args are passed, the
            bai_2d_args dictionary is also updated with the new args.
            If no args are passed, uses bai_2d_args attribute.
        """
        if not args:
            args = self.bai_2d_args
        else:
            self.bai_2d_args = args.copy()
        with self.sphere_lock:
            if self.static:
                self.bai_2d = int_2d_data_static()
            else:
                self.bai_2d = int_2d_data()

            for arch in self.arches:
                arch.integrate_2d(global_mask=self.global_mask, **args)
                self.arches[arch.idx] = arch
                self._update_bai_2d(arch)

    def _update_bai_1d(self, arch):
        """helper function to update overall bai variables.
        """
        with self.sphere_lock:
            try:
                assert list(self.bai_1d.raw.shape) == list(arch.int_1d.raw.shape)
            except (AssertionError, AttributeError):
                if not self.static:
                    self.bai_1d.raw = np.zeros(arch.int_1d.raw.shape)
                    self.bai_1d.pcount = np.zeros(arch.int_1d.pcount.shape)
                self.bai_1d.norm = np.zeros(arch.int_1d.norm.shape)
                self.bai_1d.sigma = np.zeros(arch.int_1d.norm.shape)
                self.bai_1d.sigma_raw = np.zeros(arch.int_1d.norm.shape)
            try:
                self.bai_1d += arch.int_1d
                self.bai_1d.ttheta = arch.int_1d.ttheta
                self.bai_1d.q = arch.int_1d.q
            except AttributeError:
                pass
            self.save_bai_1d()

    def _update_bai_2d(self, arch):
        """helper function to update overall bai variables.
        """
        with self.sphere_lock:
            try:
                # assert self.bai_2d.raw.shape == arch.int_2d.raw.shape
                if self.static:
                    assert self.bai_2d.i_qChi.shape == arch.int_2d.i_qChi.shape
                else:
                    assert self.bai_2d.norm.shape == arch.int_2d.norm.shape
            except (AssertionError, AttributeError):
                if self.static:
                    self.bai_2d.i_qChi = np.zeros(arch.int_2d.i_qChi.shape)
                    self.bai_2d.i_tthChi = np.zeros(arch.int_2d.i_tthChi.shape)
                    self.bai_2d.i_QxyQz = np.zeros(arch.int_2d.i_QxyQz.shape)
                else:
                    self.bai_2d.norm = np.zeros(arch.int_2d.norm.shape)
                    self.bai_2d.raw = np.zeros(arch.int_2d.raw.shape)
                    self.bai_2d.pcount = np.zeros(arch.int_2d.pcount.shape)
                    self.bai_2d.sigma = np.zeros(arch.int_2d.norm.shape)
                    self.bai_2d.sigma_raw = np.zeros(arch.int_2d.norm.shape)
            try:
                self.bai_2d.ttheta = arch.int_2d.ttheta
                self.bai_2d.q = arch.int_2d.q
                self.bai_2d.chi = arch.int_2d.chi
                if not self.static:
                    self.bai_2d += arch.int_2d
                else:
                    self.bai_2d.i_qChi += arch.int_2d.i_qChi
                    self.bai_2d.i_tthChi += arch.int_2d.i_tthChi
                    self.bai_2d.i_QxyQz += arch.int_2d.i_QxyQz
                    self.bai_2d.qz = arch.int_2d.qz
                    self.bai_2d.qxy = arch.int_2d.qxy
            except AttributeError:
                pass
            self.save_bai_2d()

    def set_multi_geo(self, **args):
        """Sets the MultiGeometry instance stored in the arch.

        args: see pyFAI.multiple_geometry.MultiGeometry. If no args are
            passed, uses mg_args attribute. Otherwise, updates
            mg_args and uses passed arguments.
        """
        self.mg_args.update(args)
        with self.sphere_lock:
            self.multi_geo = MultiGeometry(
                [a.integrator for a in self.arches], **self.mg_args
            )

    def multigeometry_integrate_1d(self, monitor=None, **kwargs):
        """Wrapper for integrate1d method of MultiGeometry.

        args:
            monitor: channel with normalization value
            kwargs: see MultiGeometry.integrate1d

        returns:
            result: result from MultiGeometry.integrate1d
        """
        with self.sphere_lock:
            lst_mask = [a.get_mask() for a in self.arches]
            if monitor is None:
                try:
                    result = self.multi_geo.integrate1d(
                        [a.map_norm for a in self.arches], lst_mask=lst_mask,
                        **kwargs
                    )
                except Exception as e:
                    print(e)
                    print('Exception 1')
                    result = self.multi_geo.integrate1d(
                        [a.map_raw for a in self.arches], lst_mask=lst_mask,
                        **kwargs
                    )
            else:
                result = self.multi_geo.integrate1d(
                    [a.map_raw for a in self.arches], lst_mask=lst_mask,
                    normalization_factor=list(self.scan_data[monitor]),
                    **kwargs
                )

            self.mgi_1d.from_result(result, self.multi_geo.wavelength)
        return result

    def multigeometry_integrate_2d(self, monitor=None, **kwargs):
        """Wrapper for integrate1d method of MultiGeometry.

        args:
            monitor: channel with normalization value
            kwargs: see MultiGeometry.integrate1d

        returns:
            result: result from MultiGeometry.integrate1d
        """
        with self.sphere_lock:
            lst_mask = [a.get_mask() for a in self.arches]
            if monitor is None:
                try:
                    result = self.multi_geo.integrate2d(
                        [a.map_raw / a.map_norm for a in self.arches], lst_mask=lst_mask,
                        **kwargs
                    )
                except Exception as e:
                    print(e)
                    print('Exception 2')
                    result = self.multi_geo.integrate2d(
                        [a.map_raw for a in self.arches], lst_mask=lst_mask,
                        **kwargs
                    )
            else:
                result = self.multi_geo.integrate2d(
                    [a.map_raw for a in self.arches], lst_mask=lst_mask,
                    normalization_factor=list(self.scan_data[monitor]),
                    **kwargs
                )

            self.mgi_2d.from_result(result, self.multi_geo.wavelength)
        return result

    def save_to_h5(self, replace=False, *args, **kwargs):
        """Saves data to hdf5 file.

        args:
            replace: bool, if True file is truncated prior to writing
                data.
            arches: list, list of arch ids to save. Deprecated.
            data_onle: bool, if true only saves the scan_data attribute
                and does not save mg_args, bai_1d_args, or bai_2d_args.
            compression: str, what compression algorithm to pass to
                h5py. See h5py documentation for acceptable compression
                algorithms.
        """
        if replace:
            mode = 'w'
        else:
            mode = 'a'
        with self.file_lock:
            with utils.catch_h5py_file(self.data_file, mode) as file:
                self._save_to_h5(file, *args, **kwargs)

    def _save_to_h5(self, grp, arches=None, data_only=False,
                    compression='lzf'):
        """Actual function for saving data, run with the file open and
            holding the file_lock.
        """
        if self.static:
            compression = None
        with self.sphere_lock:
            grp.attrs['type'] = 'EwaldSphere'

            if data_only:
                lst_attr = [
                    "scan_data", "global_mask", "overall_raw",
                ]
            else:
                lst_attr = [
                    "scan_data", "global_mask", "mg_args", "bai_1d_args",
                    "bai_2d_args", "overall_raw",
                    "static", "gi", "th_mtr", "single_img", "poni_dict",
                    "series_average"
                ]
            utils.attributes_to_h5(self, grp, lst_attr,
                                   compression=compression)

            keys = ('bai_1d', 'bai_2d', 'mgi_1d', 'mgi_2d')
            if self.static:
                keys = ('bai_1d', 'bai_2d')
            for key in keys:
                if key not in grp:
                    grp.create_group(key)
            self.bai_1d.to_hdf5(grp['bai_1d'], compression)
            self.bai_2d.to_hdf5(grp['bai_2d'], compression)
            if not self.static:
                self.mgi_1d.to_hdf5(grp['mgi_1d'], compression)
                self.mgi_2d.to_hdf5(grp['mgi_2d'], compression)

    def load_from_h5(self, replace=True, mode='r', *args, **kwargs):
        """Loads data stored in hdf5 file.

        args:
            data_only: bool, if True only loads the scan_data attribute
                and does not load mg_args, bai_1d_args, or bai_2d_args.
            set_mg: bool, if True instantiates the Multigeometry
                object.
        """
        with self.file_lock:
            if replace:
                self.reset()
            with utils.catch_h5py_file(self.data_file, mode=mode) as file:
                self._load_from_h5(file, *args, **kwargs)

    def _load_from_h5(self, grp, data_only=False, set_mg=True):
        """Actual function for loading data, run with the file open and
            holding the file_lock.
        """
        with self.sphere_lock:
            if 'type' in grp.attrs:
                if grp.attrs['type'] == 'EwaldSphere':
                    for arch in grp['arches']:
                        if int(arch) not in self.arches.index:
                            self.arches.index.append(int(arch))

                    self.arches.sort_index(inplace=True)

                    if data_only:
                        lst_attr = [
                            "scan_data", "overall_raw",
                        ]
                        utils.h5_to_attributes(self, grp, lst_attr)
                    else:
                        lst_attr = [
                            "scan_data", "mg_args", "bai_1d_args",
                            "bai_2d_args", "overall_raw",
                            "static", "gi", "th_mtr", "single_img", "poni_dict",
                            "series_average"
                        ]
                        utils.h5_to_attributes(self, grp, lst_attr)
                        self._set_args(self.bai_1d_args)
                        self._set_args(self.bai_2d_args)

                        if not self.static:
                            self._set_args(self.mg_args)
                    if "global_mask" in grp:
                        utils.h5_to_attributes(self, grp, ["global_mask"])
                    else:
                        self.global_mask = None

                    self.bai_1d.from_hdf5(grp['bai_1d'])
                    self.bai_2d.from_hdf5(grp['bai_2d'])
                    if not self.static:
                        self.mgi_1d.from_hdf5(grp['mgi_1d'])
                        self.mgi_2d.from_hdf5(grp['mgi_2d'])
                        if set_mg:
                            self.set_multi_geo(**self.mg_args)

    def set_datafile(self, fname, name=None, keep_current_data=False,
                     save_args={}, load_args={}):
        """Sets the data_file. If file exists and has data, loads in the
        data. Otherwise, creates new file and resets self.

        args:
            fname: str, new data file
            name: str or None, new name. If None, name is obtained from
                fname.
            keep_current_data: bool, if True overwrites any existing
                data in the file. Otherwise, current data is either
                overwritten by data in file or deleted if no data
                exists, except any args dicts which are untouched.
            save_args: dict, arguments to be passed to save_to_h5
            load_args: dict, arguments to be passed to load_from_h5
        """
        with self.sphere_lock:
            self.data_file = fname
            if name is None:
                self.name = os.path.split(fname)[-1].split('.')[0]
            else:
                self.name = name
            if keep_current_data:
                self.save_to_h5(replace=True, **save_args)
            else:
                if os.path.exists(fname):
                    self.load_from_h5(replace=True, **load_args)
                else:
                    self.reset()
                    self.save_to_h5(replace=True, **save_args)

    def save_bai_1d(self, compression='lzf'):
        """Function to save only the bai_1d object.

        args:
            compression: str, what compression algorithm to pass to
                h5py. See h5py documentation for acceptable compression
                algorithms.
        """
        compression = None
        if self.static:
            compression = None
        with self.file_lock:
            with utils.catch_h5py_file(self.data_file, 'a') as file:
                self.bai_1d.to_hdf5(file['bai_1d'], compression=compression)

    def save_bai_2d(self, compression='lzf'):
        """Function to save only the bai_2d object.

        args:
            compression: str, what compression algorithm to pass to
                h5py. See h5py documentation for acceptable compression
                algorithms.
        """
        compression = None
        if self.static:
            compression = None
        with self.file_lock:
            with utils.catch_h5py_file(self.data_file, 'a') as file:
                self.bai_2d.to_hdf5(file['bai_2d'], compression=compression)

    def _set_args(self, args):
        """Ensures any range args are lists.
        """
        for arg in args:
            if 'range' in arg:
                if args[arg] is not None:
                    args[arg] = list(args[arg])


def get_1D_data(h5_file, arch_ids=None, static=True):
    """Loads 1D data from hdf5 file

    args:
        h5_file: hdf5 file with processed data
        arch_ids: arches whose 1D data is loaded
        static: scan type flag

    returns:
        df: Pandas dataframe with integrated 1D data
    """
    h5_file = Path(h5_file)
    scan_name = h5_file.stem
    sphere = EwaldSphere(scan_name, data_file=str(h5_file), static=static)
    sphere.load_from_h5(replace=False, mode='r')

    df1 = pd.DataFrame(columns=('idx', 'intensity', 'tth', 'q'))
    with utils.catch_h5py_file(sphere.data_file, 'r') as file:
        if arch_ids is None:
            arch_ids = sphere.arches.index

        for idx in arch_ids:
            try:
                arch = EwaldArch(idx=idx, static=sphere.static, gi=sphere.gi)
                if str(idx) not in file['arches']:
                    print("No data can be found")
                    continue
                grp = file['arches'][str(idx)]
                if 'type' in grp.attrs:
                    if grp.attrs['type'] == 'EwaldArch':
                        lst_attr = [
                            "scan_info", "ai_args",
                            "gi", "static", "poni_dict"
                        ]
                        utils.h5_to_attributes(arch, grp, lst_attr)
                        arch.int_1d.from_hdf5(grp['int_1d'])

                df1 = df1.append({
                    'idx': idx,
                    'intensity': list(arch.int_1d.norm),
                    'tth': list(arch.int_1d.ttheta),
                    'q': list(arch.int_1d.q)},
                    ignore_index=True
                )
            except KeyError:
                pass

    df1.set_index(df1['idx'], inplace=True)
    df2 = sphere.scan_data
    df2.rename_axis('idx')

    try:
        df = pd.concat([df1, df2.loc[df1.index]], axis=1, join='outer')
    except KeyError:
        df = df1
    return df.set_index(df['idx'])



from multiprocessing import shared_memory
from multiprocessing.managers import SharedMemoryManager
from multiprocessing import RLock
import traceback


def none_greater(a, b):
    if a is None:
        return False
    return a >= b


def none_lesser(a, b):
    if a is None:
        return False
    return a < b


def multiply_none(a, b, default=None):
    if a is None:
        return default
    return a * b


class DummyManager:
    """Behaves like shared_memory.Manager without invoking process

    Used to ensure consistent behavior between using a remote manager
    and using this as its own contained process.
    """
    def __init__(self):
        self._shared = []

    def SharedMemory(self, size):
        """Returns a new SharedMemory instance with the specified size in
        bytes, untracked."""
        sms = shared_memory.SharedMemory(None, create=True, size=size)
        self._shared.append(sms)
        return sms

    def ShareableList(self, sequence):
        """Returns a new ShareableList instance populated with the values
        from the input sequence, untracked."""
        sl = shared_memory.ShareableList(sequence)
        self._shared.append(sl.shm)
        return sl

    def __del__(self):
        for s in self._shared:
            try:
                s.close()
                s.unlink()
            except:
                traceback.print_exc()



import numpy as np

from ..datashare import SMBase, synced, locked
from ..datashare.smarray import bytes_to_shape, shape_to_bytes
from .int_data import parse_unit, int_1d_data


def _get_size(xsize, ysize):
    length = ysize * 5 + xsize * 2
    itemsize = np.dtype(float).itemsize
    size = int(itemsize * length)
    if size < itemsize * 2:
        size = itemsize * 2
    return size


def _get_bounds(arr: np.ndarray):
    if (arr == 0).all():
        return 0, 0
    else:
        c = np.nonzero(arr)[0]
        return c[0], len(c)


class SMIntData1D(SMBase):
    def __init__(self, addr=None, ysize=0, xsize=0, no_zeros=False, **kwargs):
        format_list = [
            '0'*128,
            int(0),
            int(0),
            int(ysize),
            int(xsize),
            int(0),
            True
        ]
        size = _get_size(xsize, ysize)
        SMBase.__init__(self, addr=addr, format_list=format_list, size=size, **kwargs)
        with self.mutex:
            if addr is None:
                self._shl[3] = ysize
                self._shl[4] = xsize
                self._shl[5] = 0
                self._shl[6] = no_zeros
            self.npview = np.ndarray(
                (self._shl[3] * 5 + self._shl[4] * 2,),
                dtype=float,
                buffer=self._shm.buf
            )
            self._set_arrays()

    def _set_arrays(self):
        for i, attr in enumerate(['raw', 'pcount', 'norm', 'sigma', 'sigma_raw']):
            super(SMBase, self).__setattr__(
                attr,
                self.npview[i*self._shl[3]:(i+1)*self._shl[3]]
            )
        for i, attr in enumerate(['ttheta', 'q']):
            super(SMBase, self).__setattr__(
                attr,
                self.npview[5*self._shl[3] + i*self._shl[4]:5*self._shl[3] + (i+1)*self._shl[4]]
            )

    def __setattr__(self, name, value):
        with self.mutex:
            self.check_memory()
            if name in ['raw', 'norm', 'pcount', 'sigma', 'sigma_raw', 'ttheta', 'q']:
                self.__dict__[name][:] = value[:]
            else:
                super(SMBase, self).__setattr__(name, value)

    @synced
    def resize(self, ysize, xsize=None):
        if xsize is None:
            _xsize = ysize
        else:
            _xsize = xsize
        size = _get_size(_xsize, ysize)
        self._recap(size)
        self._shl[3] = ysize
        self._shl[4] = xsize
        self._set_arrays()

    @synced
    def from_result(self, result, wavelength, monitor=1):
        """Parses out result obtained by pyFAI AzimuthalIntegrator.

        args:
            result: object returned by AzimuthalIntegrator
            wavelength: float, energy of the beam in meters
        """
        self.ttheta, self.q = parse_unit(
            result, wavelength)

        self.pcount = result._count
        self.raw = result._sum_signal / monitor
        self.norm = self.raw / self.pcount
        if result.sigma is None:
            self.sigma = result._sum_signal
            self.sigma = np.sqrt(self.sigma)
            self.sigma = self.sigma / (self.pcount * monitor)
            self.sigma_raw = result._sum_signal / (monitor ** 2)
        else:
            self.sigma = result.sigma / monitor
            self.sigma_raw = ((result._count * result.sigma) ** 2) / (monitor ** 2)



# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports
import os
from collections import OrderedDict
import time
import copy
import traceback

# Other imports
import numpy as np
import fabio

# Qt imports
import pyFAI
from pyqtgraph import Qt
from pyqtgraph.parametertree import ParameterTree, Parameter

# This module imports
from xdart.modules.spec import MakePONI, get_spec_header
from xdart.utils._utils import get_meta_from_spec
from xdart.utils._utils import get_mask_array, get_img_data
from xdart.utils.containers import PONI
from xdart.modules.ewald import EwaldArch, EwaldSphere
from .wrangler_widget import wranglerWidget, wranglerThread, wranglerProcess
from .specUI import Ui_Form
from ....gui_utils import NamedActionParameter
from ....widgets import MaskWidget
from xdart.utils.pyFAI_binaries import pyFAI_drawmask_main
from xdart.utils.pyFAI_binaries import MaskImageWidgetXdart

from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

QFileDialog = Qt.QtWidgets.QFileDialog

DETECTOR_DICT = {
    "Pilatus100k": {
        "shape": (195, 487)
    }
}

params = [
    {'name': 'Scan Number', 'type': 'int', 'value': 1},
    {'name': 'Spec File', 'type': 'str', 'default': ''},
    NamedActionParameter(name='spec_file_browse', title= 'Browse...'),
    {'name': 'Image Directory', 'type': 'str', 'default': ''},
    NamedActionParameter(name='image_dir_browse', title= 'Browse...'),
    {'name': 'Calibration PONI File', 'type': 'str', 'default': ''},
    NamedActionParameter(name='poni_file_browse', title= 'Browse...'),
    {'name': 'Timeout', 'type': 'float', 'value': 5},
    {'name': 'Rotation Motors', 'type': 'group', 'children': [
        {'name': 'Rot1', 'type': 'str', 'default': ''},
        {'name': 'Rot2', 'type': 'str', 'default': 'tth', 'value': 'tth'},
        {'name': 'Rot3', 'type': 'str', 'default': ''}
    ]},
    {'name': 'Calibration Angles', 'type': 'group', 'children': [
        {'name': 'Rot1', 'type': 'float', 'value': 0},
        {'name': 'Rot2', 'type': 'float', 'value': 0},
        {'name': 'Rot3', 'type': 'float', 'value': 0}
    ]},
    {'name': 'Detector', 'type': 'list', 'values': [
            "Pilatus100k"
        ],
     'value':'Pilatus100k'},
    {'name': 'det_orientation', 'title': 'Detector Orientation', 'type': 'list',
     'values': [0, 90, 180, 270], 'value': 0},
    NamedActionParameter(name='set_mask', title= 'Make mask...'),
    {'name': 'mask_file', 'title': 'Mask File', 'type': 'str', 'default': ''},
    NamedActionParameter(name='mask_file_browse', title='Browse...'),
]


class specWrangler(wranglerWidget):
    """Widget for integrating data associated with spec file. Can be
    used "live", will continue to poll data folders until image data
    and corresponding spec data are available.
    
    attributes:
        command_queue: Queue, used to send commands to thread
        file_lock, mp.Condition, process safe lock for file access
        fname: str, path to data file
        parameters: pyqtgraph Parameter, stores parameters from user
        scan_name: str, current scan name, used to handle syncing data
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        thread: wranglerThread or subclass, QThread for controlling
            processes
        timeout: int, how long before thread stops looking for new
            data.
        tree: pyqtgraph ParameterTree, stores and organizes parameters
        ui: Ui_Form from qtdesigner
    
    methods:
        cont, pause, stop: functions to pass continue, pause, and stop
            commands to thread via command_queue 
        enabled: Enables or disables interactivity
        set_image_dir: sets the image directory
        set_poni_file: sets the calibration poni file
        set_spec_file: sets the spec data file
        set_fname: Method to safely change file name
        setup: Syncs thread parameters prior to starting
    
    signals:
        finished: Connected to thread.finished signal
        sigStart: Tells tthetaWidget to start the thread and prepare
            for new data.
        sigUpdateData: int, signals a new arch has been added.
        sigUpdateFile: (str, str), sends new scan_name and file name
            to tthetaWidget.
        showLabel: str, connected to thread showLabel signal, sets text
            in specLabel
    """
    showLabel = Qt.QtCore.Signal(str)

    def __init__(self, fname, file_lock, parent=None):
        """fname: str, file path
        file_lock: mp.Condition, process safe lock
        """
        # ic()
        super().__init__(fname, file_lock, parent)

        # Setup gui elements
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.ui.startButton.clicked.connect(self.sigStart.emit)
        self.ui.pauseButton.clicked.connect(self.pause)
        self.ui.stopButton.clicked.connect(self.stop)
        self.ui.continueButton.clicked.connect(self.cont)

        # Setup parameter tree
        self.tree = ParameterTree()
        self.parameters = Parameter.create(
            name='spec_wrangler', type='group', children=params
        )
        self.tree.setParameters(self.parameters, showTop=False)
        self.layout = Qt.QtWidgets.QVBoxLayout(self.ui.paramFrame)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.addWidget(self.tree)
        
        # Wire signals from parameter tree based buttons
        self.parameters.child('spec_file_browse').sigActivated.connect(
            self.set_spec_file
        )
        self.parameters.child('image_dir_browse').sigActivated.connect(
            self.set_image_dir
        )
        self.parameters.child('poni_file_browse').sigActivated.connect(
            self.set_poni_file
        )
        self.parameters.child('mask_file').sigValueChanged.connect(
            self.update_mask
        )
        self.parameters.child('mask_file_browse').sigActivated.connect(
            self.set_mask_file
        )
        self.parameters.child('det_orientation').sigValueChanged.connect(
            self.set_det_orientation
        )
        
        # Set attributes
        self.scan_number = self.parameters.child('Scan Number').value()
        self.timeout = self.parameters.child('Timeout').value()
        self.parameters.sigTreeStateChanged.connect(self.setup)
        self.det_orientation = int(self.parameters.child('det_orientation').value())
        
        # Setup thread
        self.thread = specThread(
            self.command_queue, 
            self.sphere_args, 
            self.fname, 
            self.file_lock, 
            self.scan_name, 
            0,
            {},
            {},
            None,
            5,
            self.mask,
            self.det_orientation,
            self
        )
        self.thread.showLabel.connect(self.ui.specLabel.setText)
        self.thread.sigUpdateFile.connect(self.sigUpdateFile.emit)
        self.thread.finished.connect(self.finished.emit)
        self.thread.sigUpdate.connect(self.sigUpdateData.emit)

        # self.mask_widget = MaskWidget()
        detector_name = self.parameters.child("Detector").value()
        self.detector = pyFAI.detector_factory(detector_name)

        # data = np.zeros(DETECTOR_DICT[key]["shape"])
        # data[0, 0] = 1
        # self.mask_widget.set_data(data.T)
        # self.mask_widget.hide()

        self.mask_file = self.parameters.child('mask_file').value()
        self.mask = get_mask_array(self.detector, self.mask_file)
        self.thread.mask = self.mask

        self.parameters.child('set_mask').sigActivated.connect(
            self.run_pyfai_drawmask
            # self.launch_mask_widget
        )

        self.mask_window = None
        # self.mask_widget.newMask.connect(self.set_mask)

        self.setup()

    def setup(self):
        """Sets up the child thread, syncs all parameters.
        """
        # ic()
        self.thread.mp_inputs.update(self._get_mp_inputs())
        lsf_inputs = self._get_lsf_inputs()
        self.thread.lsf_inputs.update(lsf_inputs)
        self.scan_number = self.parameters.child('Scan Number').value()
        self.scan_name = lsf_inputs['spec_file_name'] + '_scan' + \
                         str(self.scan_number)
        self.fname = os.path.join(lsf_inputs['spec_file_path'],
                                  self.scan_name + '.hdf5')
        self.thread.fname = self.fname
        self.thread.scan_name = self.scan_name
        self.thread.scan_number = self.scan_number
        self.thread.img_dir = self.parameters.child('Image Directory').value()
        self.timeout = self.parameters.child('Timeout').value()
        self.thread.timeout = self.parameters.child('Timeout').value()
        self.thread.file_lock = self.file_lock
        self.thread.sphere_args = self.sphere_args
        self.thread.mask = self.mask
        self.thread.det_orientation = self.det_orientation

        if self.parameters.child('Rotation Motors').child('Rot2').value() == '':
            self.parameters.child('Rotation Motors').child('Rot2').setValue('tth')

    def pause(self):
        if self.thread.isRunning():
            self.command_queue.put('pause')

    def cont(self):
        if self.thread.isRunning():
            self.command_queue.put('continue')

    def stop(self):
        if self.thread.isRunning():
            self.command_queue.put('stop')
    
    def set_spec_file(self):
        """Opens file dialogue and sets the spec data file
        """
        # ic()
        fname, _ = Qt.QtWidgets.QFileDialog().getOpenFileName()
        if fname != '':
            self.parameters.child('Spec File').setValue(fname)
    
    def set_image_dir(self):
        """Opens file dialogue and sets the image directory
        """
        # ic()
        dname = Qt.QtWidgets.QFileDialog.getExistingDirectory(self)
        if dname != '':
            self.parameters.child('Image Directory').setValue(dname)
    
    def set_poni_file(self):
        """Opens file dialogue and sets the calibration file
        """
        # ic()
        # fname, _ = Qt.QtWidgets.QFileDialog().getOpenFileName()
        fname, _ = Qt.QtWidgets.QFileDialog().getOpenFileName(
            filter="PONI (*.poni *.PONI)"
        )
        if fname != '':
            self.parameters.child('Calibration PONI File').setValue(fname)

    def set_det_orientation(self):
        self.det_orientation = int(self.parameters.child('det_orientation').value())
        self.update_mask()

    def _get_mp_inputs(self):
        """Organizes inputs for MakePONI from parameters.
        """
        # ic()
        mp_inputs = OrderedDict(
            rotations = {
                "rot1": None,
                "rot2": 'tth',
                "rot3": None
            },
            calib_rotations = {
                "rot1": 0,
                "rot2": 0,
                "rot3": 0
            },
            poni_file = None,
            spec_dict = {}
        )
        rot_mot = self.parameters.child('Rotation Motors')
        for child in rot_mot:
            if child.value() == "":
                mp_inputs['rotations'][child.name().lower()] = None
            else:
                mp_inputs['rotations'][child.name().lower()] = child.value()
        
        cal_rot = self.parameters.child('Calibration Angles')
        for child in cal_rot:
            if child.value() == 0:
                pass
            else:
                mp_inputs['calib_rotations'][child.name().lower()] = child.value()
        
        mp_inputs['poni_file'] = self.parameters.child(
                                     'Calibration PONI File').value()

        return mp_inputs

    def _get_lsf_inputs(self):
        """Organizes inputs for LoadSpecFile from parameters. No longer
        used.
        """
        # ic()
        dirname, fname = os.path.split(self.parameters.child('Spec File').value())
        lsf_inputs = OrderedDict(
            spec_file_path=dirname,
            spec_file_name=fname
        )

        return lsf_inputs

    def enabled(self, enable):
        """Sets tree and start button to enable.
        
        args:
            enable: bool, True for enabled False for disabled.
        """
        # ic()
        self.tree.setEnabled(enable)
        self.ui.startButton.setEnabled(enable)

    def set_mask_file(self):
        """Opens file dialogue and sets the mask file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter="EDF (*.edf)"
        )
        if fname != '':
            self.parameters.child('mask_file').setValue(fname)
            self.mask_file = fname
        self.update_mask()

    def update_mask(self):
        self.mask_file = self.parameters.child('mask_file').value()
        self.mask = get_mask_array(self.detector, self.mask_file, det_orientation=self.det_orientation)

        ic('updating mask...')
        ic(self.mask, self.mask_file, self.detector, self.det_orientation)

    # def set_mask(self, idx, mask):
    #     self.mask = np.arange(self.mask_widget.data.size)[mask.ravel() == 1]
    #     self.thread.mask = self.mask

    # def launch_mask_widget(self):
    #     key = self.parameters.child("Detector").value()
    #     data = np.zeros(DETECTOR_DICT[key]["shape"])
    #     data[0, 0] = 1
    #     if self.mask is not None:
    #         _mask = np.zeros_like(data.T)
    #         _mask.ravel()[self.mask] = 1
    #         self.mask_widget.set_data(data.T, base=_mask)
    #     else:
    #         self.mask_widget.set_data(data.T)
    #     self.mask_widget.show()

    def run_pyfai_drawmask(self):
        filters = f'Images (*.tif *.tiff *.raw)'
        processFile, _ = QFileDialog().getOpenFileName(
            filter=filters,
            caption='Choose Image File',
            options=QFileDialog.DontUseNativeDialog
        )
        if not os.path.exists(processFile):
            print('No Image Chosen')
            return

        self.mask_window = MaskImageWidgetXdart()
        self.mask_window.setWindowModality(Qt.QtCore.Qt.WindowModal)
        self.mask_window.show()

        try:
            image = fabio.open(processFile).data
        except OSError:
            image = get_img_data(processFile, self.detector)

        pyFAI_drawmask_main(self.mask_window, image, processFile)

        self.parameters.child('mask_file').setValue(self.mask_window.outFile)
        self.mask_file = self.mask_window.outFile


class specThread(wranglerThread):
    """Thread for controlling the specProcessor process. Receives
    manages a command and signal queue to pass commands from the main
    thread and communicate back relevant signals.
    
    attributes:
        command_q: mp.Queue, queue to send commands to process
        file_lock: mp.Condition, process safe lock for file access
        fname: str, path to data file.
        img_dir: str, path to image directory
        lsf_inputs: dict, input parameters for LoadSpecFile
        mp_inputs: dict, input parameters for MakePONI
        scan_name: str, name of current scan
        scan_number: int, number of current scan
        input_q: mp.Queue, queue for commands sent from parent
        signal_q: mp.Queue, queue for commands sent from process
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        timeout: float or int, how long to continue checking for new
            data.
        mask: 1D array, indices of 2D array to be masked
        det_orientation: float or int, Orientation of detector

    signals:
        showLabel: str, sends out text to be used in specLabel
    
    methods:
        run: Main method, called by start
    """
    showLabel = Qt.QtCore.Signal(str)

    def __init__(
            self, 
            command_queue, 
            sphere_args, 
            fname, 
            file_lock,
            scan_name, 
            scan_number,
            mp_inputs, 
            lsf_inputs,
            img_dir, 
            timeout,
            mask,
            det_orientation,
            parent=None):
        """command_queue: mp.Queue, queue for commands sent from parent
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        fname: str, path to data file.
        file_lock: mp.Condition, process safe lock for file access
        scan_name: str, name of current scan
        scan_number: int, number of current scan
        mp_inputs: dict, input parameters for MakePONI
        lsf_inputs: dict, input parameters for LoadSpecFile
        img_dir: str, path to image directory
        timeout: float or int, how long to continue checking for new
            data.
        mask: 1D array, indices of 2D array to be masked
        det_orientation: float or int, Orientation of detector
        """
        # ic()
        super().__init__(command_queue, sphere_args, fname, file_lock, parent)
        self.scan_name = scan_name
        self.scan_number = scan_number
        self.mp_inputs = mp_inputs
        self.lsf_inputs = lsf_inputs
        self.img_dir = img_dir
        self.timeout = timeout
        self.mask = mask
        self.det_orientation = det_orientation

    def run(self):
        """Initializes specProcess and watches for new commands from
        parent or signals from the process.
        """
        # ic()
        process = specProcess(
            self.command_q, 
            self.signal_q, 
            self.sphere_args, 
            self.scan_name,
            self.scan_number,
            self.fname, 
            self.file_lock,
            self.lsf_inputs, 
            self.mp_inputs,
            self.img_dir,
            self.timeout,
            self.mask,
            self.det_orientation
        )
        process.start()
        last = False
        # Main loop
        while True:
            # Check for new commands
            if not self.input_q.empty():
                command = self.input_q.get()
                print(command)
                self.command_q.put(command)
            
            #Check for new updates
            if not self.signal_q.empty():
                signal, data = self.signal_q.get()
                if signal == 'update':
                    self.sigUpdate.emit(data)
                elif signal == 'message':
                    self.showLabel.emit(data)
                elif signal == 'new_scan':
                    self.sigUpdateFile.emit(self.scan_name, self.fname)
                elif signal == 'TERMINATE':
                    last = True
            
            # Breaks on signal from process
            if last:
                break
        
        # Empty queues of any other items after main loop ends.
        self._empty_q(self.signal_q)
        self._empty_q(self.command_q)
        process.join()
    
    def _empty_q(self, q):
        """Empties out a given queue.
        args:
            q: Queue
        """
        # ic()
        while not q.empty():
            _ = q.get()
    

class specProcess(wranglerProcess):
    """Process for integrating scanning area detector data. Checks for
    a specified scan in a spec file, and then searches for associated
    raw files. Data is stored with an EwaldSphere object, saving all
    data to an hdf5 file.
    
    attributes:
        command_q: mp.Queue, queue for commands from parent thread.
        file_lock: mp.Condition, process safe lock for file access
        fname: str, path to data file
        img_dir: str, path to image directory
        lsf_inputs: dict, input parameters for LoadSpecFile
        mp_inputs: dict, input parameters for MakePONI
        scan_name: str, name of current scan
        scan_number: int, number of current scan
        signal_q: queue to place signals back to parent thread.
        spec_name: str, spec base name used for checking raw files
        specFile: dict, data from spec file read into dataframes
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        timeout: float or int, how long to continue checking for new
            data.
        user: str, user name from spec file
    
    methods:
        _main: Controls flow of integration, checking for commands,
            providing updates, and catching errors.
        read_raw: method for reading in binary .raw files and returning
            data as a numpy array.
        wrangle: Method which handles data loading from files.
    """
    def __init__(self, command_q, signal_q, sphere_args, scan_name, 
                 scan_number, fname, file_lock, lsf_inputs, mp_inputs,
                 img_dir, timeout, mask, det_orientation, *args, **kwargs):
        """command_q: mp.Queue, queue for commands from parent thread.
        signal_q: queue to place signals back to parent thread.
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        scan_name: str, name of current scan
        scan_number: int, number of current scan
        fname: str, path to data file
        file_lock: mp.Condition, process safe lock for file access
        lsf_inputs: dict, input parameters for LoadSpecFile
        mp_inputs: dict, input parameters for MakePONI
        img_dir: str, path to image directory
        timeout: float or int, how long to continue checking for new
            data.
        det_orientation: float or int, Orientation of detector
        """
        # ic()
        super().__init__(command_q, signal_q, sphere_args, fname, file_lock,
                         *args, **kwargs)
        self.lsf_inputs = lsf_inputs
        self.mp_inputs = mp_inputs
        self.scan_name = scan_name
        self.scan_number = scan_number
        self.img_dir = img_dir
        self.specFile = OrderedDict(
            header=dict(
                    meta={},
                    motors={},
                    motors_r={},
                    detectors={},
                    detectors_r={}
                    ),
            scans={},
            scans_meta={}
        )
        self.user = None
        self.spec_name = None
        self.timeout = timeout
        self.mask = mask
        self.det_orientation = det_orientation

    def _main(self):
        """Checks for commands in queue, sends back updates through
        signal queue, and catches errors. Calls wrangle method for
        reading in data, then performs integration.
        """
        # ic()
        # Initialize sphere and save to disk, send update for new scan
        sphere = EwaldSphere(self.scan_name, data_file=self.fname,
                             global_mask=self.mask,
                             **self.sphere_args)
        sphere.global_mask = self.mask
        with self.file_lock:
            sphere.save_to_h5(replace=True)
            self.signal_q.put(('new_scan', None))
        
        # Operation instantiated within process to avoid conflicts with locks
        make_poni = MakePONI()
        make_poni.inputs.update(self.mp_inputs)

        # full spec path grabbed from lsf_inputs
        # TODO: just give full spec path as argument
        spec_path = os.path.join(self.lsf_inputs['spec_file_path'],
                                 self.lsf_inputs['spec_file_name'])

        # Enter main loop
        i = 0
        pause = False
        start = time.time()
        while True:
            # Check for commands, or wait if paused
            if not self.command_q.empty() or pause:
                command = self.command_q.get()
                print(command)
                if command == 'stop':
                    break
                elif command == 'continue':
                    pause = False
                elif command == 'pause':
                    pause = True
                    continue

            # ic(i, spec_path)
            # Get result from wrangle
            try:
                flag, data = self.wrangle(i, spec_path, make_poni)
            # Errors associated with image not yet taken
            except (KeyError, FileNotFoundError, AttributeError, ValueError):
                elapsed = time.time() - start
                if elapsed > self.timeout:
                    self.signal_q.put(('message', "Timeout occurred"))
                    self.signal_q.put(('TERMINATE', None))
                    break
                else:
                    continue
            start = time.time()
            
            # Unpack data and load into sphere
            # TODO: Test how long integrating vs io takes
            if flag == 'image':
                idx, map_raw, scan_info, poni = data
                arch = EwaldArch(
                    idx, map_raw, PONI.from_yamdict(poni), scan_info=scan_info
                )
                
                # integrate image to 1d and 2d arrays
                arch.integrate_1d(global_mask=self.mask, **sphere.bai_1d_args)
                arch.integrate_2d(global_mask=self.mask, **sphere.bai_2d_args)

                # Add arch copy to sphere, save to file
                with self.file_lock:
                    sphere.add_arch(
                        arch=arch.copy(), calculate=False, update=True,
                        get_sd=True, set_mg=False
                    )
                    sphere.save_to_h5(data_only=True, replace=False)
                
                self.signal_q.put(('message', f'Image {i} integrated'))
                self.signal_q.put(('update', idx))
                i += 1
            
            # Check if terminate signal sent
            elif flag == 'TERMINATE' and data is None:
                self.signal_q.put(('TERMINATE', None))
                break
        
        # If loop ends, signal terminate to parent thread.
        self.signal_q.put(('TERMINATE', None))

    def wrangle(self, i, spec_path, make_poni):
        """Method for reading in data from raw files and spec file.
        
        args:
            i: int, index of image to check
            spec_path: str, absolute path to spec file
            make_poni: MakePONI, operation for creating poni objects
        
        returns:
            flag: str, signal for what kind of data to expect.
            data: tuple (int, numpy array, dict, dict), the
                index of the data, raw image array, metadata, and PONI
                dict associated with the image.
        """
        # ic()
        self.signal_q.put(('message', f'Checking for {i}'))

        # reads in spec data file header
        self.specFile['header'] = get_spec_header(spec_path)

        # checks for user and spec_name information
        if self.user is None:
            self.user = self.specFile['header']['meta']['User']
        if self.spec_name is None:
            self.spec_name = os.path.basename(self.specFile['header']['meta']['File'][0])

        # Construct raw_file path from attributes and index
        raw_file = self._get_raw_path(i)

        Counters, Motors, Extras = get_meta_from_spec(raw_file, spec_file=spec_path)
        image_meta = Motors | Counters
        # ic(image_meta)

        # reads in scan data
        # self.specFile['scans'][self.scan_number], \
        #     self.specFile['scans_meta'][self.scan_number] = get_spec_scan(
        #     spec_path, self.scan_number, self.specFile['header']
        # )

        # Get scan meta data
        # if self.scan_number in self.specFile['scans'].keys():
        #     image_meta = self.specFile['scans']\
        #                         [self.scan_number].loc[i].to_dict()
        #
        # else:
        #     image_meta = self.specFile['current_scan'].loc[i].to_dict()
        # ic(image_meta)

        # Get poni dict based on meta data
        make_poni.inputs['spec_dict'] = copy.deepcopy(image_meta)
        poni = copy.deepcopy(make_poni.run())
        # ic(poni)
        
        # Read raw file into numpy array
        # ic(raw_file)
        arr = self.read_raw(raw_file)

        self.signal_q.put(('message', f'Image {i} wrangled'))

        return 'image', (i, arr, image_meta, poni)

    def _get_raw_path(self, i):
        """Creates raw path name from attributes, following spec
        convention.
        
        args:
            i: int, index of image
        
        returns:
            raw_file: str, absolute path to .raw image file.
        """
        # ic()
        im_base = '_'.join([
            self.user,
            self.spec_name,
            'scan' + str(self.scan_number),
            str(i).zfill(4)
        ])
        return os.path.join(self.img_dir, im_base + '.raw')
    
    def read_raw(self, file, mask=True):
        """Reads in .raw file and returns a numpy array.
        
        args:
            file: str, path to .raw file
            mask: bool, if True clips the edges of the image
        
        returns:
            map_raw: numpy array, image data.
        """
        # ic()
        with open(file, 'rb') as im:
            # ic('file opened')
            arr = np.fromstring(im.read(), dtype='int32')
            # ic('converted to np')
            arr = arr.reshape((195, 487))
            if mask:
                for i in range(0, 10):
                    arr[:, i] = -2.0
                for i in range(477, 487):
                    arr[:, i] = -2.0

            if self.det_orientation == 0:
                arr = arr.T
            elif self.det_orientation == 90:
                arr = arr
            elif self.det_orientation == 180:
                arr = arr.T[::-1, :]
            elif self.det_orientation == 270:
                arr = arr[::-1, :]
            return arr



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Users\walroth\Documents\repos\xdart\xdart\experiments\ttheta_scan\gui\h5viewerUI.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(666, 445)
        self.layout = QtWidgets.QGridLayout(Form)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setVerticalSpacing(0)
        self.layout.setObjectName("layout")
        self.frame = QtWidgets.QFrame(Form)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.layout.addWidget(self.frame, 0, 0, 1, 1)
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.layout.addWidget(self.label, 1, 0, 1, 1)
        self.listScans = QtWidgets.QListWidget(Form)
        self.listScans.setObjectName("listScans")
        self.layout.addWidget(self.listScans, 2, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.layout.addWidget(self.label_2, 1, 1, 1, 1)
        self.listData = QtWidgets.QListWidget(Form)
        self.listData.setObjectName("listData")
        self.layout.addWidget(self.listData, 2, 1, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Scans"))
        self.label_2.setText(_translate("Form", "Data"))




from collections import namedtuple
import struct
from sys import getsizeof


def _get_type_size(fchar, sample):
    """
    Returns the length of a list of bytes to represent the character
    using struct.pack

    Parameters
    ----------
    fchar : str
        One letter code for the struct format
    sample : variable
        A valid representation of the type to be packed

    Returns
    -------
    int
        Number of bytes needed to pack the type
    """
    return len(list(struct.pack(fchar, sample)))


DataType = namedtuple("DataType", ["size", "equivalent", "unpack", "pack"])


char_t = DataType(_get_type_size('b', 0), int,
                  lambda x: struct.unpack('b', x)[0],
                  lambda x: struct.pack('b', x))
uchar_t = DataType(_get_type_size('B', 0), int,
                   lambda x: struct.unpack('B', x)[0],
                   lambda x: struct.pack('B', x))
bool_t = DataType(_get_type_size('?', True), bool,
                  lambda x: struct.unpack('?', x)[0],
                  lambda x: struct.pack('?', x))
short_t = DataType(_get_type_size('h', 0), int,
                   lambda x: struct.unpack('h', x)[0],
                   lambda x: struct.pack('h', x))
ushort_t = DataType(_get_type_size('H', 0), int,
                    lambda x: struct.unpack('H', x)[0],
                    lambda x: struct.pack('H', x))
int_t = DataType(_get_type_size('i', 0), int,
                 lambda x: struct.unpack('i', x)[0],
                 lambda x: struct.pack('i', x))
uint_t = DataType(_get_type_size('I', 0), int,
                  lambda x: struct.unpack('I', x)[0],
                  lambda x: struct.pack('I', x))
long_t = DataType(_get_type_size('l', 0), int,
                  lambda x: struct.unpack('l', x)[0],
                  lambda x: struct.pack('l', x))
ulong_t = DataType(_get_type_size('L', 0), int,
                   lambda x: struct.unpack('L', x)[0],
                   lambda x: struct.pack('L', x))
llong_t = DataType(_get_type_size('q', 0), int,
                   lambda x: struct.unpack('q', x)[0],
                   lambda x: struct.pack('q', x))
ullong_t = DataType(_get_type_size('Q', 0), int,
                    lambda x: struct.unpack('Q', x)[0],
                    lambda x: struct.pack('Q', x))
float_t = DataType(_get_type_size('f', 0), float,
                   lambda x: struct.unpack('f', x)[0],
                   lambda x: struct.pack('f', x))
double_t = DataType(_get_type_size('d', 0), float,
                    lambda x: struct.unpack('d', x)[0],
                    lambda x: struct.pack('d', x))


key_to_dtype = {
    "char_t": char_t,
    "uchar_t": uchar_t,
    "bool_t": bool_t,
    "short_t": short_t,
    "ushort_t": ushort_t,
    "int_t": int_t,
    "uint_t": uint_t,
    "long_t": long_t,
    "ulong_t": ulong_t,
    "llong_t": llong_t,
    "ullong_t": ullong_t,
    "float_t": float_t,
    "double_t": double_t,
}

dtype_to_key = {val: key for key, val in key_to_dtype.items()}



import yaml
import logging
from logging import info, warning, critical, debug


def load_yaml_config(config_filename):
    """Load the data structure from the YAML config file."""
    try:
        with open(config_filename, 'r') as stream:
            rawConfig = "".join(stream.readlines())

        parsedConfig = next(yaml.load_all(rawConfig))
    except IOError:
        critical("Could not open {} YAML config file for motor variables")
        raise
    else:
        return parsedConfig


def save_yaml_config(data, config_filename):
    """Load the data structure from the YAML config file."""
    try:
        with open(config_filename, 'w') as stream:
            yaml.dump(data, stream)  # Write a YAML representation of data
    except IOError:
        critical("Could not write {} YAML config file for motor variables".format(config_filename))
        raise

    return


